#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
YOLO model definition for detection.
"""

import pickle

import numpy as np

from keras import Model

from cnn2snn import load_quantized_model, quantize

from ..layer_blocks import yolo_head_block
from ..imagenet.model_akidanet import akidanet_imagenet
from ..utils import fetch_file


def yolo_base(input_shape=(224, 224, 3),
              classes=1,
              nb_box=5,
              alpha=1.0,
              weight_quantization=0,
              activ_quantization=0,
              input_weight_quantization=None,
              input_scaling=(127.5, -1)):
    """ Instantiates the YOLOv2 architecture.

    Args:
        input_shape (tuple): input shape tuple
        classes (int): number of classes to classify images into
        nb_box (int): number of anchors boxes to use
        alpha (float): controls the width of the model
        weight_quantization (int): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization: sets all activations in the model to have a
            particular activation quantization bitwidth.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization: sets weight quantization in the first layer.
            Defaults to weight_quantization value.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_scaling (tuple, optional): scale factor and offset to apply to
            inputs. Defaults to (127.5, -1). Note that following Akida
            convention, the scale factor is a number used as a divider.

    Returns:
        keras.Model: a Keras Model instance.

    """
    # Overrides input weight quantization if None
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Create an AkidaNet network without top layers
    base_model = akidanet_imagenet(
        input_shape=input_shape,
        alpha=alpha,
        include_top=False,
        weight_quantization=weight_quantization,
        activ_quantization=activ_quantization,
        input_weight_quantization=input_weight_quantization,
        input_scaling=input_scaling)

    # Add YOLO top layers to the base model
    input_shape = base_model.input_shape
    x = yolo_head_block(base_model.layers[-1].output, num_boxes=nb_box, classes=classes)
    model = Model(inputs=base_model.input, outputs=x, name='yolo_base')

    # Initialize detection layer weights
    layer = model.get_layer("detection_layer")
    detection_weights = layer.get_weights()

    mu, sigma = 0, 0.1

    grid_size = model.output_shape[1:3]
    grid_area = grid_size[0] * grid_size[1]

    dw_kernel = np.random.normal(mu, sigma,
                                 size=detection_weights[0].shape) / grid_area
    pw_kernel = np.random.normal(mu, sigma,
                                 size=detection_weights[1].shape) / grid_area
    bias = np.random.normal(mu, sigma,
                            size=detection_weights[2].shape) / grid_area

    layer.set_weights([dw_kernel, pw_kernel, bias])

    if ((weight_quantization != 0) or (activ_quantization != 0) or
            (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def yolo_widerface_pretrained():
    """
    Helper method to retrieve a `yolo_base` model that was trained on WiderFace
    dataset and the anchors that are needed to interpet the model output.

    Returns:
        keras.Model, list: a Keras Model instance and a list of anchors.

    """
    anchors_name = 'widerface_anchors.pkl'
    anchors_path = fetch_file(
        'http://data.brainchip.com/dataset-mirror/widerface/' + anchors_name,
        fname=anchors_name,
        file_hash='325f92336a310d83fed71765436ee343bf3e39cbc12fd099d30677761aee9376',
        cache_subdir='datasets/widerface')
    with open(anchors_path, 'rb') as handle:
        anchors = pickle.load(handle)

    model_name = 'yolo_akidanet_widerface_iq8_wq4_aq4.h5'
    file_hash = 'd55744cbfbbe1131aa26015f38d99f1df7026347bae8f66683259e366e7b6e03'
    model_path = fetch_file('http://data.brainchip.com/models/AkidaV1/yolo/' + model_name,
                            fname=model_name,
                            file_hash=file_hash,
                            cache_subdir='models')

    return load_quantized_model(model_path), anchors


def yolo_voc_pretrained():
    """
    Helper method to retrieve a `yolo_base` model that was trained on PASCAL
    VOC2012 dataset for 'person' and 'car' classes only, and the anchors that
    are needed to interpet the model output.

    Returns:
        keras.Model, list: a Keras Model instance and a list of anchors.

    """
    anchors_name = 'voc_anchors.pkl'
    anchors_path = fetch_file(
        'http://data.brainchip.com/dataset-mirror/voc/' + anchors_name,
        fname=anchors_name,
        file_hash='b1fe1ed12691e100646cf52b1320f05abd17b2f546d3e12cdee87758cc9ed0ba',
        cache_subdir='datasets/voc')
    with open(anchors_path, 'rb') as handle:
        anchors = pickle.load(handle)

    model_name = 'yolo_akidanet_voc_iq8_wq4_aq4.h5'
    file_hash = 'e65b0b6bd4b08c2796c3bbea89343748195e29e240ce28e70489c53d06ca69d9'
    model_path = fetch_file('http://data.brainchip.com/models/AkidaV1/yolo/' + model_name,
                            fname=model_name,
                            file_hash=file_hash,
                            cache_subdir='models')

    return load_quantized_model(model_path), anchors



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
ImageNet1K dataset mean and std constants definition.
"""

import warnings


# Taken from timm.data.constants
IMAGENET_MEAN = (0.485, 0.456, 0.406)
IMAGENET_STD = (0.229, 0.224, 0.225)


def obtain_input_shape(input_shape, default_size, min_size, include_top):
    """Internal utility to compute/validate a model's input shape.

    Args:
        input_shape: either None (will return the default model input shape),
            or a user-provided shape to be validated.
        default_size: default input width/height for the model.
        min_size: minimum input width/height accepted by the model.
        include_top: whether the model is expected to
            be linked to a classifier via a Flatten layer.

    Returns:
        tuple of integers: input shape (may include None entries).

    Raises:
        ValueError: in case of invalid argument values.
    """
    if input_shape and len(input_shape) == 3:
        if input_shape[-1] not in {1, 3}:
            warnings.warn('This model usually expects 1 or 3 input channels. '
                          'However, it was passed an input_shape with ' +
                          str(input_shape[-1]) + ' input channels.')
        default_shape = (default_size, default_size, input_shape[-1])
    else:
        default_shape = (default_size, default_size, 3)

    if input_shape:
        if input_shape is not None:
            if len(input_shape) != 3:
                raise ValueError(
                    '`input_shape` must be a tuple of three integers.')
            if ((input_shape[0] is not None and input_shape[0] < min_size) or
                    (input_shape[1] is not None and input_shape[1] < min_size)):
                raise ValueError('Input size must be at least ' +
                                 str(min_size) + 'x' + str(min_size) +
                                 '; got `input_shape=' + str(input_shape) + '`')
    else:
        if include_top:
            input_shape = default_shape
        else:
            input_shape = (None, None, 3)
    if include_top:
        if None in input_shape:
            raise ValueError('If `include_top` is True, '
                             'you should specify a static `input_shape`. '
                             'Got `input_shape=' + str(input_shape) + '`')
    return input_shape



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Portrait128 training script.
"""

import os
import argparse

import numpy as np
from keras.preprocessing.image import ImageDataGenerator
from keras.metrics import BinaryIoU

from cnn2snn import load_quantized_model

from akida_models.param_scheduler import get_cosine_lr_scheduler

from ..training import (get_training_parser, compile_model, evaluate_model, print_history_stats,
                        RestoreBest, save_model)


def get_data(path, batch_size):
    """ Loads Portrait128 data.

    Args:
        path (str): path to npy data
        batch_size (int): the batch size

    Returns:
        tuple: train generator, validation generator, steps per epoch and validation step
    """
    # Load the dataset
    x_train = np.load(os.path.join(path, "img_uint8.npy"))
    y_train = np.load(os.path.join(path, "msk_uint8.npy")) / 255

    # Data generator for training and validation
    data_gen_args = dict(width_shift_range=0.1,
                         height_shift_range=0.1,
                         zoom_range=0.2,
                         horizontal_flip=True,
                         validation_split=0.2)

    image_datagen = ImageDataGenerator(**data_gen_args)
    mask_datagen = ImageDataGenerator(**data_gen_args)

    seed = 1
    train_image_generator = image_datagen.flow(x_train,
                                               batch_size=batch_size,
                                               shuffle=True,
                                               subset='training',
                                               seed=seed)

    train_mask_generator = mask_datagen.flow(y_train,
                                             batch_size=batch_size,
                                             shuffle=True,
                                             subset='training',
                                             seed=seed)

    val_image_generator = image_datagen.flow(x_train,
                                             batch_size=batch_size,
                                             shuffle=True,
                                             subset='validation',
                                             seed=seed)

    val_mask_generator = mask_datagen.flow(y_train,
                                           batch_size=batch_size,
                                           shuffle=True,
                                           subset='validation',
                                           seed=seed)

    # Combine generators into one which yields image and masks
    train_generator = zip(train_image_generator, train_mask_generator)
    val_generator = zip(val_image_generator, val_mask_generator)

    return (train_generator, val_generator, train_image_generator.n // batch_size,
            val_image_generator.n // batch_size)


def train_model(model, train_gen, steps_per_epoch, val_gen, val_steps, epochs, learning_rate):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        train_gen (keras.ImageDataGenerator): train data generator
        steps_per_epoch (int): training steps
        val_gen (keras.ImageDataGenerator): validation data generator
        val_steps (int): validation steps
        epochs (int): the number of epochs
        learning_rate (float): the learning rate
    """
    # Define learning rate scheduler
    callbacks = [get_cosine_lr_scheduler(learning_rate, epochs * steps_per_epoch, True)]

    # Model checkpoints (save best model and retrieve it when training is complete)
    restore_model = RestoreBest(model, monitor="val_binary_io_u")
    callbacks.append(restore_model)

    history = model.fit(train_gen,
                        epochs=epochs,
                        steps_per_epoch=steps_per_epoch,
                        validation_steps=val_steps,
                        validation_data=val_gen,
                        callbacks=callbacks)
    print_history_stats(history)


def main():
    """ Entry point for script and CLI usage.
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument("-d", "--data", type=str,
                               default="/hdd/datasets/Portrait128/data/",
                               help="Path to the Portrait128 data.")

    parsers = get_training_parser(batch_size=32, tune=True, global_parser=global_parser)
    args = parsers[0].parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # Compile model
    learning_rate = 3e-5
    if args.action == "tune":
        learning_rate /= 10

    compile_model(model, learning_rate=learning_rate, loss='binary_crossentropy',
                  metrics=[BinaryIoU(), 'accuracy'])

    # Load data
    train_gen, val_gen, steps_per_epoch, val_steps = get_data(args.data, args.batch_size)

    # Disable QuantizeML assertions
    os.environ["ASSERT_ENABLED"] = "0"

    # Train model
    if args.action in ["train", "tune"]:
        train_model(model, train_gen, steps_per_epoch, val_gen,
                    val_steps, args.epochs, learning_rate)
        save_model(model, args.model, args.savemodel, args.action)
    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            raise NotImplementedError('Converting segmentation models is not supported yet.')
        evaluate_model(model, val_gen, steps=val_steps, print_history=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for YOLO models.
"""

import os
import pickle
import argparse
import numpy as np

from keras import Model
from keras.layers import Reshape
from keras.callbacks import EarlyStopping

from cnn2snn import load_quantized_model, convert

from .yolo_loss import YoloLoss
from .map_evaluation import MapEvaluation
from .batch_generator import BatchYoloGenerator
from ..training import get_training_parser, freeze_model_before, compile_model, save_model


def get_data(data_path, anchors_path):
    """ Loads data.

    Args:
        data_path (str): path to data pickle file
        anchors_path (str): path to anchors pickle file

    Returns:
        dict, dict, list, list: train and validation data, labels and anchors
    """
    with open(data_path, 'rb') as handle:
        train_data, valid_data, labels = pickle.load(handle)

    with open(anchors_path, 'rb') as handle:
        anchors = pickle.load(handle)

    return train_data, valid_data, labels, anchors


def train(model, train_data, valid_data, anchors, labels, obj_threshold,
          nms_threshold, epochs, batch_size, grid_size):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        train_data (dict): training data
        valid_data (dict): validation data
        anchors (list): list of anchors
        labels (list): list of labels
        obj_threshold (float): confidence threshold for a box
        nms_threshold (float): non-maximal supression threshold
        epochs (int): the number of epochs
        batch_size (int): the batch size
        grid_size (tuple): the grid size
    """
    TRAIN_TIMES = 10

    # Build batch generators
    input_shape = model.input.shape[1:]

    train_generator = BatchYoloGenerator(input_shape=input_shape,
                                         data=train_data,
                                         grid_size=grid_size,
                                         labels=labels,
                                         anchors=anchors,
                                         batch_size=batch_size)

    valid_generator = BatchYoloGenerator(input_shape=input_shape,
                                         data=valid_data,
                                         grid_size=grid_size,
                                         labels=labels,
                                         anchors=anchors,
                                         batch_size=batch_size,
                                         jitter=False)

    # Create callbacks
    early_stop_cb = EarlyStopping(monitor='val_loss',
                                  min_delta=0.001,
                                  patience=10,
                                  mode='min',
                                  verbose=1)

    map_evaluator_cb = MapEvaluation(model,
                                     valid_data,
                                     labels,
                                     anchors,
                                     period=4,
                                     obj_threshold=obj_threshold,
                                     nms_threshold=nms_threshold)

    callbacks = [early_stop_cb, map_evaluator_cb]

    # Start the training process
    model.fit(x=train_generator,
              steps_per_epoch=len(train_generator) * TRAIN_TIMES,
              epochs=epochs,
              validation_data=valid_generator,
              callbacks=callbacks,
              workers=12,
              max_queue_size=40)


def evaluate(model, valid_data, anchors, labels, obj_threshold, nms_threshold):
    """ Evaluates model performances.

    Args:
        model (keras.Model or akida.Model): the model to evaluate
        valid_data (dict): evaluation data
        anchors (list): list of anchors
        labels (list): list of labels
        obj_threshold (float): confidence threshold for a box
        nms_threshold (float): non-maximal supression threshold
    """
    # Create the mAP evaluator
    map_evaluator = MapEvaluation(model,
                                  valid_data,
                                  labels,
                                  anchors,
                                  obj_threshold=obj_threshold,
                                  nms_threshold=nms_threshold,
                                  is_keras_model=isinstance(model, Model))

    # Compute mAP scores and display them
    mAP, average_precisions = map_evaluator.evaluate_map()
    for label, average_precision in average_precisions.items():
        print(labels[label], '{:.4f}'.format(average_precision))
    print('mAP: {:.4f}'.format(mAP))


def extract_samples(im_size, data, grid_size, anchors, labels, num_samples, out_file):
    """ Extract samples from data and save them to a .npz file.

    Args:
        im_size (int): image size
        data (dict): data container
        grid_size (tuple): the grid size
        anchors (list): list of anchors
        labels (list): list of labels
        num_samples (int): number of samples to extract
        out_file (str): name of output file
    """
    data_generator = BatchYoloGenerator(input_shape=im_size,
                                        data=data,
                                        grid_size=grid_size,
                                        labels=labels,
                                        anchors=anchors,
                                        batch_size=num_samples)
    np.savez(out_file, data_generator[0][0])
    print(f"Samples saved as {out_file}")


def main():
    """ Entry point for script and CLI usage.
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument(
        "-d",
        "--data",
        default=None,
        help="The pickle file generated by the preprocessing"
        " step that contains data variables")
    global_parser.add_argument("-ap",
                               "--anchors_path",
                               default=None,
                               help="Path to anchors boxes file.")
    global_parser.add_argument("-obj",
                               "--obj_thresh",
                               type=float,
                               default=0.5,
                               help="Confidence threshold for a box")
    global_parser.add_argument("-nms",
                               "--nms_thresh",
                               type=float,
                               default=0.5,
                               help="Non-Maximal Suppression threshold.")

    parsers = get_training_parser(batch_size=128,
                                  freeze_before=True,
                                  tune=True,
                                  global_parser=global_parser)
    subparser = parsers[-1]
    extract_parser = subparser.add_parser("extract",
                                          help="Extract samples",
                                          parents=[global_parser])
    extract_parser.add_argument("-o", "--out_file", type=str,
                                default='widerface_samples.npz', help="Name of the output file.")

    args = parsers[0].parse_args()

    # Load data
    train_data, valid_data, labels, anchors = get_data(args.data,
                                                       args.anchors_path)

    # Load the source model
    base_model = load_quantized_model(args.model)

    # Create a final reshape layer for loss computation
    grid_size = base_model.output_shape[1:3]
    output = Reshape(
        (grid_size[1], grid_size[0], len(anchors), 4 + 1 + len(labels)),
        name="YOLO_output")(base_model.output)

    # Build the full model
    model = Model(base_model.input, output)

    # Freeze the model
    if "freeze_before" in args:
        freeze_model_before(model, args.freeze_before)

    # Compile model
    learning_rate = 1e-4
    if args.action == "tune":
        learning_rate = 1e-5

    compile_model(model,
                  learning_rate=learning_rate,
                  loss=YoloLoss(anchors, grid_size, args.batch_size),
                  metrics=None)

    # Disable QuantizeML assertions
    os.environ["ASSERT_ENABLED"] = "0"

    # Train model
    if args.action in ["train", "tune"]:
        train(model, train_data, valid_data, anchors, labels, args.obj_thresh,
              args.nms_thresh, args.epochs, args.batch_size, grid_size)
        # Remove the last reshape layer introduced for training
        model = Model(model.input, model.layers[-2].output)
        save_model(model, args.model, args.savemodel, args.action)

    elif args.action == 'eval':
        # Evaluate model accuracy
        if args.akida:
            # Drop the last reshape layer that is not Akida compatible
            if model.layers[-1].name == 'YOLO_output':
                model = Model(model.input, model.layers[-2].output)
            model = convert(model)
        evaluate(model, valid_data, anchors, labels, args.obj_thresh,
                 args.nms_thresh)

    elif args.action == 'extract':
        input_shape = model.input.shape[1:]
        extract_samples(input_shape, train_data, grid_size, anchors,
                        labels, args.batch_size, args.out_file)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
AkidaNet model definition for ImageNet classification.

AkidaNet is an NSoC optimized model inspired from VGG and MobileNet V1
architectures. It can be used for multiple use cases through transfer learning.

"""

from keras import Model, regularizers
from keras.layers import Input, Dropout, Rescaling

from cnn2snn import quantize, load_quantized_model

from .imagenet_utils import obtain_input_shape
from ..layer_blocks import conv_block, separable_conv_block, dense_block
from ..utils import fetch_file

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/AkidaV1/akidanet/'


def akidanet_imagenet(input_shape=None,
                      alpha=1.0,
                      include_top=True,
                      pooling=None,
                      classes=1000,
                      weight_quantization=0,
                      activ_quantization=0,
                      input_weight_quantization=None,
                      input_scaling=(128, -1)):
    """Instantiates the AkidaNet architecture.

    Args:
        input_shape (tuple, optional): shape tuple. Defaults to None.
        alpha (float, optional): controls the width of the model.
            Defaults to 1.0.

            * If `alpha` < 1.0, proportionally decreases the number of filters
              in each layer.
            * If `alpha` > 1.0, proportionally increases the number of filters
              in each layer.
            * If `alpha` = 1, default number of filters from the paper are used
              at each layer.
        include_top (bool, optional): whether to include the fully-connected
            layer at the top of the model. Defaults to True.
        pooling (str, optional): optional pooling mode for feature extraction
            when `include_top` is `False`.
            Defaults to None.

            * `None` means that the output of the model will be the 4D tensor
              output of the last convolutional block.
            * `avg` means that global average pooling will be applied to the
              output of the last convolutional block, and thus the output of the
              model will be a 2D tensor.
        classes (int, optional): optional number of classes to classify images
            into, only to be specified if `include_top` is `True`. Defaults to 1000.
        weight_quantization (int, optional): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.
            Defaults to 0.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization (int, optional): sets all activations in the model to have a
            particular activation quantization bitwidth.
            Defaults to 0.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization (int, optional): sets weight quantization in the first layer.
            Defaults to weight_quantization value.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_scaling (tuple, optional): scale factor and offset to apply to
            inputs. Defaults to (128, -1). Note that following Akida convention,
            the scale factor is an integer used as a divider.

    Returns:
        keras.Model: a Keras model for AkidaNet/ImageNet.

    Raises:
        ValueError: in case of invalid input shape.
    """
    # check if overrides have been provided and override
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Define weight regularization, will apply to the convolutional layers and
    # to all pointwise weights of separable convolutional layers.
    weight_regularizer = regularizers.l2(4e-5)

    # Determine proper input shape and default size.
    if input_shape is None:
        default_size = 224
    else:
        rows = input_shape[0]
        cols = input_shape[1]

        if rows == cols and rows in [128, 160, 192, 224]:
            default_size = rows
        else:
            default_size = 224

    input_shape = obtain_input_shape(input_shape,
                                     default_size=default_size,
                                     min_size=32,
                                     include_top=include_top)

    img_input = Input(shape=input_shape, name="input")

    if input_scaling is None:
        x = img_input
    else:
        scale, offset = input_scaling
        x = Rescaling(1. / scale, offset, name="rescaling")(img_input)

    x = conv_block(x,
                   filters=int(32 * alpha),
                   name='conv_0',
                   kernel_size=(3, 3),
                   padding='same',
                   use_bias=False,
                   strides=2,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    x = conv_block(x,
                   filters=int(64 * alpha),
                   name='conv_1',
                   kernel_size=(3, 3),
                   padding='same',
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    x = conv_block(x,
                   filters=int(128 * alpha),
                   name='conv_2',
                   kernel_size=(3, 3),
                   padding='same',
                   strides=2,
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    x = conv_block(x,
                   filters=int(128 * alpha),
                   name='conv_3',
                   kernel_size=(3, 3),
                   padding='same',
                   use_bias=False,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(256 * alpha),
                             name='separable_4',
                             kernel_size=(3, 3),
                             padding='same',
                             strides=2,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(256 * alpha),
                             name='separable_5',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_6',
                             kernel_size=(3, 3),
                             padding='same',
                             strides=2,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_7',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_8',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_9',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_10',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(512 * alpha),
                             name='separable_11',
                             kernel_size=(3, 3),
                             padding='same',
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    x = separable_conv_block(x,
                             filters=int(1024 * alpha),
                             name='separable_12',
                             kernel_size=(3, 3),
                             padding='same',
                             strides=2,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    # Last separable layer with global pooling
    layer_pooling = 'global_avg' if include_top or pooling == 'avg' else None
    x = separable_conv_block(x,
                             filters=int(1024 * alpha),
                             name='separable_13',
                             kernel_size=(3, 3),
                             padding='same',
                             pooling=layer_pooling,
                             use_bias=False,
                             add_batchnorm=True,
                             add_activation=True,
                             pointwise_regularizer=weight_regularizer)

    if include_top:
        x = Dropout(1e-3, name='dropout')(x)
        x = dense_block(x,
                        classes,
                        name='classifier',
                        add_batchnorm=False,
                        add_activation=False,
                        kernel_regularizer=weight_regularizer)

    # Create model.
    model = Model(img_input,
                  x,
                  name='akidanet_%0.2f_%s_%s' %
                  (alpha, input_shape[0], classes))

    if ((weight_quantization != 0) or (activ_quantization != 0) or
            (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def akidanet_imagenet_pretrained(alpha=1.0):
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    ImageNet dataset.

    Args:
        alpha (float, optional): width of the model, allowed values in [0.25,
            0.5, 1]. Defaults to 1.0.

    Returns:
        keras.Model: a Keras Model instance.

    """
    if alpha == 1.0:
        model_name = 'akidanet_imagenet_224_iq8_wq4_aq4.h5'
        file_hash = '359e0ff05abbb26fe215830ca467ef40761767c0d77f8c743c6d6e4fffa7a925'
    elif alpha == 0.5:
        model_name = 'akidanet_imagenet_224_alpha_50_iq8_wq4_aq4.h5'
        file_hash = '1d9493115d43625f2644f8265f71b8487c4019047fc331e892a233a3d6520371'
    elif alpha == 0.25:
        model_name = 'akidanet_imagenet_224_alpha_25_iq8_wq4_aq4.h5'
        file_hash = '9146d6228d859d8b8db1c1b7a795471096e5a49ee4ff5656eabc6be749d42f5e'
    else:
        raise ValueError(
            f"Requested model with alpha={alpha} is not available.")

    model_path = fetch_file(BASE_WEIGHT_PATH + model_name,
                            fname=model_name,
                            file_hash=file_hash,
                            cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_faceidentification_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    CASIA Webface dataset and that performs face identification.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_faceidentification_iq8_wq4_aq4.h5'
    file_hash = '117df1118124c5d2cc0cf867dcc10fa1dfbc3978c1bfe0dcba727ac369e48765'
    model_path = fetch_file(BASE_WEIGHT_PATH + model_name,
                            fname=model_name,
                            file_hash=file_hash,
                            cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_plantvillage_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    PlantVillage dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_plantvillage_iq8_wq4_aq4.h5'
    file_hash = '1400910c774fd78a5e6ea227ff28cb28e79ecec0909a378068cd9f40ddaf4e0a'
    model_path = fetch_file(BASE_WEIGHT_PATH + model_name,
                            fname=model_name,
                            file_hash=file_hash,
                            cache_subdir='models')
    return load_quantized_model(model_path)


def akidanet_vww_pretrained():
    """
    Helper method to retrieve an `akidanet_imagenet` model that was trained on
    VWW dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet_vww_iq8_wq4_aq4.h5'
    file_hash = 'cd130d90ed736447b6244dc1228e708b9dab20af0d2bf57b9a49df4362467ea8'
    model_path = fetch_file(BASE_WEIGHT_PATH + model_name,
                            fname=model_name,
                            file_hash=file_hash,
                            cache_subdir='models')
    return load_quantized_model(model_path)



# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Preprocessing tools for ImageNet dataset.
"""

import numpy as np
import keras
import tensorflow as tf
import tensorflow_addons as tfa

from akida_models.imagenet.imagenet_labels2names import imagenet_labels


class RandomColorJitter(keras.layers.Layer):
    """RandomColorJitter class.

    Randomly adds color jitter to an image. Color jitter means to add random brightness, contrast,
    saturation, and hue to an image. There is a 80% chance that an image will be randomly
    color-jittered. Taken on https://keras.io/examples/vision/barlow_twins/

    Args:
        proba(float, optional): Probability of applying the color jitter. Defaults to 0.8.
    """

    def __init__(self, *args, proba=0.8, **kwargs):
        super().__init__(*args, **kwargs)
        self.proba = proba

    @tf.function
    def call(self, image):
        def _color_jitter(image):
            image = tf.image.random_brightness(image, 0.8)
            image = tf.image.random_contrast(image, 0.4, 1.6)
            image = tf.image.random_saturation(image, 0.4, 1.6)
            image = tf.image.random_hue(image, 0.2)
            return image
        return tf.cond(
            tf.random.uniform([]) < self.proba, lambda: _color_jitter(image), lambda: image)


class ThreeAugment(keras.layers.Layer):
    """Define a simple data augmentation pipeline of three augmentations, following the explaining
    in the paper: https://arxiv.org/abs/2204.07118. For that, this augmentation choses one of:
        - GrayScale: This favors color invariance and give more focus on shapes.
        - Solarization: This adds strong noise on the colour to be more robust to the variation
          of colour intensity and so focus more on shape
        - Gaussian Blur: In order to slightly alter details in the image.
    """
    @tf.function
    def call(self, image):
        def _to_gray():
            return tf.image.grayscale_to_rgb(tf.image.rgb_to_grayscale(image))

        def _solarize():
            # Taken of: https://keras.io/examples/vision/barlow_twins/
            return tf.where(image < 10, image, 255 - image)

        def _gaussian_blur():
            # Taken of: https://keras.io/examples/vision/barlow_twins/
            s = np.random.random()
            return tfa.image.gaussian_filter2d(image=image, sigma=s)

        proba = tf.random.uniform([])
        cases = [(proba < 1 / 3, _to_gray), (proba < 2 / 3, _solarize)]
        return tf.case(cases, default=_gaussian_blur, name='3-Augment', exclusive=False)


DATA_AUGMENTATION = keras.Sequential([RandomColorJitter(), ThreeAugment()])


@tf.function
def preprocess_image(image, image_size, training=False, data_aug=None):
    """ ImageNet data preprocessing.

    Preprocessing includes cropping, and resizing for both training and
    validation images. Training preprocessing introduces some random distortion
    of the image to improve accuracy.

    Args:
        image (tf.Tensor): input image as a 3-D tensor
        image_size (int): desired image size
        training (bool, optional): True for training preprocessing, False for
            validation and inference. Defaults to False.
        data_aug (keras.Sequential, optional): data augmentation. Defaults to None.

    Returns:
        :obj:`tensorflow.Tensor`: preprocessed image
    """
    shape = tf.shape(image)

    if training:
        # For training: crop, flip and resize
        bbox_begin, bbox_size, _ = tf.image.sample_distorted_bounding_box(
            shape,
            tf.zeros([0, 0, 4], tf.float32),  # force using whole image
            use_image_if_no_bounding_boxes=True,
            min_object_covered=0.1,
            aspect_ratio_range=[0.75, 1.33],
            area_range=[0.05, 1.0],
            max_attempts=100)

        image = tf.slice(image, bbox_begin, bbox_size)
        image = tf.image.resize(image, [image_size, image_size])

        # Make all data augmentation after resize to decrease computational cost
        image = tf.image.random_flip_left_right(image)
        if data_aug is not None:
            image = data_aug(image)
    else:
        # For validation/inference: aspect preserving resize and central crop
        height = tf.cast(shape[0], tf.float32)
        width = tf.cast(shape[1], tf.float32)

        # Scale image before cropping, keeping aspect ratio
        resize_min = np.round(image_size * 1.143).astype(np.float32)
        scale_ratio = resize_min / tf.minimum(height, width)

        # Convert back to int for TF ops
        new_height = tf.cast(height * scale_ratio, tf.int32)
        new_width = tf.cast(width * scale_ratio, tf.int32)

        image = tf.image.resize(image, [new_height, new_width])

        # Second: central crop to desired image_size
        image = tf.image.resize_with_crop_or_pad(image, image_size, image_size)

    return tf.cast(image, tf.float32)


def index_to_label(index):
    """ Function to get an ImageNet label from an index.

    Args:
        index: between 0 and 999

    Returns:
        str: a string of comma separated labels
    """
    return imagenet_labels[index]



#!/usr/bin/env python
# *****************************************************************************
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Tools for Knowledge Distillation training.

Originated from https://keras.io/examples/vision/knowledge_distillation/.

Reference Hinton et al. (2015) https://arxiv.org/abs/1503.02531
"""

import tensorflow as tf
from functools import partial
from tensorflow import GradientTape
from keras import Model
from keras.losses import KLDivergence, CategoricalCrossentropy


class Distiller(Model):
    """ The class that will be used to train the student model using the
    distillation knowledge method.

    Reference `Hinton et al. (2015) <https://arxiv.org/abs/1503.02531>`_.

    Args:
        student (keras.Model): the student model
        teacher (keras.Model): the well trained teacher model
        alpha (float, optional): weight to student_loss_fn and 1-alpha
            to distillation_loss_fn. Defaults to 0.1
    """

    def __init__(self, student, teacher, alpha=0.1):
        super().__init__()
        self.teacher = teacher
        self.student = student
        self.student_loss_fn = None
        self.distillation_loss_fn = None
        self.alpha = alpha

    @property
    def base_model(self):
        return self.student

    @property
    def layers(self):
        return self.base_model.layers

    def compile(self,
                optimizer,
                metrics,
                student_loss_fn,
                distillation_loss_fn):
        """ Configure the distiller.

        Args:
            optimizer (keras.optimizers.Optimizer): Keras optimizer
                for the student weights
            metrics (keras.metrics.Metric): Keras metrics for evaluation
            student_loss_fn (keras.losses.Loss): loss function of difference
                between student predictions and ground-truth
            distillation_loss_fn (keras.losses.Loss): loss function of
                difference between student predictions and teacher predictions
        """

        super().compile(optimizer=optimizer, metrics=metrics)
        self.student_loss_fn = student_loss_fn
        self.distillation_loss_fn = distillation_loss_fn

    def train_step(self, data):
        # Unpack data
        x, y = data

        # Forward pass of teacher
        teacher_predictions = self.teacher(x, training=False)
        with GradientTape() as tape:
            # Forward pass of student
            student_predictions = self.student(x, training=True)
            # Compute losses
            student_loss = self.student_loss_fn(y, student_predictions)
            distillation_loss = self.distillation_loss_fn(
                teacher_predictions, student_predictions)
            loss = self.alpha * student_loss + (1 -
                                                self.alpha) * distillation_loss

        # Compute gradients
        trainable_vars = self.student.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)

        # Update weights
        self.optimizer.apply_gradients(zip(gradients, trainable_vars))

        # Update the metrics configured in `compile()`.
        self.compiled_metrics.update_state(y, student_predictions)

        # Return a dict of performance
        results = {m.name: m.result() for m in self.metrics}
        results.update({
            "student_loss": student_loss,
            "distillation_loss": distillation_loss
        })
        return results

    def test_step(self, data):
        # Unpack the data
        x, y = data

        # Compute predictions
        y_prediction = self.student(x, training=False)

        # Calculate the loss
        student_loss = self.student_loss_fn(y, y_prediction)

        # Update the metrics.
        self.compiled_metrics.update_state(y, y_prediction)

        # Return a dict of performance
        results = {m.name: m.result() for m in self.metrics}
        results.update({"student_loss": student_loss})
        return results

    def save(self, *args, **kwargs):
        return self.base_model.save(*args, **kwargs)

    def save_weights(self, *args, **kwargs):
        return self.base_model.save_weights(*args, **kwargs)

    def load_weights(self, *args, **kwargs):
        return self.base_model.load_weights(*args, **kwargs)


class DeitDistiller(Distiller):
    """Distiller class to train the student model using the
    Knowledge Distillation (KD) method, found on https://arxiv.org/pdf/2012.12877.pdf

    The main difference with the classic KD is that the student has to produce two potential
    classification outputs. This type of training is based on the assumption that each output
    has sufficiently interacted with the whole model, therefore the main architecture can be
    trained through two different sources, as follows:

        >>> output, output_kd = student(input)
        >>> output_tc = teacher(input)
        >>> student_loss = student_loss_fn(y_true, output)
        >>> distillation_loss = distillation_loss_fn(output_tc, output_kd)

    This means we will expect to have different inputs for each loss, unlike classical KD,
    where the student's prediction is shared for both losses. However, given that each
    classifier has interacted with the student model, the gradient of each loss will contribute
    to the update of the model weights according to the alpha percentage.

    Args:
        student (keras.Model): the student model
        teacher (keras.Model): the well trained teacher model
        alpha (float, optional): weight to student_loss_fn and 1-alpha
            to distillation_loss_fn. Defaults to 0.1
        temperature (float, optional): if ``distiller_type`` when compile is equal to 'soft',
            this value will be used as temperature parameter of KLDistillationLoss.
            Defaults to 1.0.
    """

    def __init__(self, student, *args, temperature=1.0, **kwargs):
        assert len(student.outputs) == 2, "Student must be a multi-output model, with 2 outputs"

        # Append an output with the sum of heads
        y = tf.math.add_n(student.outputs) / 2
        _student = Model(student.inputs, student.outputs + [y], name=student.name)

        super().__init__(_student, *args, **kwargs)
        self._student = student
        self.temperature = temperature

    @property
    def base_model(self):
        return self._student

    def compile(self, optimizer, metrics, student_loss_fn, distiller_type):
        """ Configure the distiller.

        Args:
            optimizer (keras.optimizers.Optimizer): Keras optimizer
                for the student weights
            metrics (keras.metrics.Metric): Keras metrics for evaluation
            student_loss_fn (keras.losses.Loss): loss function of difference
                between student predictions and ground-truth
            distiller_type (str): loss function type to define the difference
                between student predictions and teacher-truth, within ['soft', 'hard', 'none'] which
                will result in performing KLDistillationLoss, CategoricalCrossentropy or
                student_loss_fn only respectively.
        """
        assert distiller_type in ['soft', 'hard', 'none']

        def _loss_forward(y_true, y_pred, loss_fn, index=0, **kwargs):
            if isinstance(y_pred, (tuple, list)):
                y_pred = y_pred[index]
            return loss_fn(y_true, y_pred, **kwargs)

        def _compile_distillation_loss_fn():
            if distiller_type == "soft":
                distillation_loss_fn = KLDistillationLoss(temperature=self.temperature)
            else:
                # Follow https://arxiv.org/pdf/2012.12877.pdf, this variant takes the
                # hard decision of the teacher as a true label. Therefore, we add the
                # prediction encoder, as well as a label smoothing equal to 0.1
                y = tf.math.softmax(self.teacher.outputs[0], axis=-1)
                self.teacher = Model(self.teacher.inputs, y, name=self.teacher.name)
                distillation_loss_fn = CategoricalCrossentropy(
                    from_logits=True, label_smoothing=0.1)
            return partial(_loss_forward, loss_fn=distillation_loss_fn, index=1)

        if distiller_type == "none" or self.teacher is None:
            # In this case, we just train the first output of student
            self.teacher = distillation_loss_fn = None
            self.student = Model(self.student.inputs,
                                 self.student.outputs[0], name=self.student.name)
            self.student.compile(optimizer, student_loss_fn, metrics)
        else:
            distillation_loss_fn = _compile_distillation_loss_fn()
            student_loss_fn = partial(_loss_forward, loss_fn=student_loss_fn)
            super().compile(optimizer, metrics, student_loss_fn, distillation_loss_fn)

    def _update_metrics(self, metrics):
        # Rename keys in the result dictionary for a more explicit display
        return {k.replace('output_1_', 'head_').replace('output_2_', 'dist_head_')
                 .replace('output_3_', ''): v for k, v in metrics.items()}

    def train_step(self, data):
        if self.teacher is None:
            return self.student.train_step(data)
        return self._update_metrics(super().train_step(data))

    def test_step(self, data):
        if self.teacher is None:
            return self.student.test_step(data)
        return self._update_metrics(super().test_step(data))


class KLDistillationLoss(KLDivergence):
    """
    The `KLDistillationLoss` is a simple wrapper around the KLDivergence loss
    that accepts raw predictions instead of probability distributions.

    Before invoking the KLDivergence loss, it converts the inputs predictions to
    probabilities by dividing them by a constant 'temperature' and applies a
    softmax.

    Args:
        temperature (float): temperature for softening probability
            distributions. Larger temperature gives softer distributions.
    """

    def __init__(self, temperature=3):
        self.temperature = temperature
        super().__init__()

    def __call__(self, y_true, y_pred, sample_weight=None):
        # Following https://github.com/facebookresearch/deit/blob/main/losses.py#L63
        # The result of KLDivergence must be scaled
        scale_factor = tf.constant(self.temperature ** 2, dtype=tf.float32)
        return super().__call__(
            tf.nn.softmax(y_true / self.temperature, axis=1),
            tf.nn.softmax(y_pred / self.temperature, axis=1)) * scale_factor



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
This module defines a custom loss function for CenterNet training
"""
from keras import backend as K
import tensorflow as tf


class CenternetLoss():
    """ Computes CenterNet loss from a model raw output.

    Args:
        alpha (float, optional): alpha parameter in heatmap loss. Defaults to 2.0.
        gamma (float, optional): gamma parameter in heatmap loss. Defaults to 4.0.
        eps (float, optional): epsilon parameter in heatmap loss. Defaults to 1e-12.
        heatmap_loss_weight (float, optional): heatmap loss weight. Defaults to 1.0.
        wh_loss_weight (float, optional): location loss weight. Defaults to 0.1.
        offset_loss_weight (float, optional): offset loss weight. Defaults to 1.0.

    """

    def __init__(self,
                 alpha=2.0,
                 gamma=4.0,
                 eps=1e-12,
                 heatmap_loss_weight=1.0,
                 wh_loss_weight=0.1,
                 offset_loss_weight=1.0):
        # Parameters for the gaussian focal loss for the heatmap branch
        self._alpha = alpha
        self._gamma = gamma
        self._eps = eps

        # Loss weight parameters
        self.heatmap_loss_weight = heatmap_loss_weight
        self.wh_loss_weight = wh_loss_weight
        self.offset_loss_weight = offset_loss_weight

    def _transform_netout(self, y_pred_raw):
        """Transforms the output of the network:

        - cast to float32
        - extracts the // wh, offset and heatmap from fused map if necessary
        - applies sigmoid to the heatmap prediction

        Args:
            y_pred_raw (tf.Tensor): raw network predictions.

        Returns:
            tuple of tf.Tensor: Predictions transformed on xy, wh and offset values.
        """
        y_pred_raw = tf.cast(y_pred_raw, dtype=tf.float32)
        y_pred_xy = K.sigmoid(y_pred_raw[..., :-4])
        y_pred_wh = y_pred_raw[..., -4:-2]
        y_pred_offset = y_pred_raw[..., -2:]
        return y_pred_xy, y_pred_wh, y_pred_offset

    def _get_targets(self, y_true):
        """Extract ground truth for each branch and compute avg_factor, wh_offset_target_weight
        here so we don't have to pass it through the whole model

        Args:
            y_true (tf.Tensor): ground truth.

        Returns:
            tuple of tf.Tensor: labels in xy, wh, offset, avg_factor and wh_offset format.
        """
        target_xy = y_true[..., :-4]
        target_wh = y_true[..., -4:-2]
        target_offset = y_true[..., -2:]

        # Extract the average factor counts the number of targets to be learned
        # max(1, center_heatmap_target.eq(1).sum())
        tmp = tf.equal(tf.constant(1.0, dtype=y_true.dtype), target_xy)
        tmp = tf.cast(tmp, dtype=tf.float32)
        tmp = tf.reduce_sum(tmp)
        avg_factor = tf.reduce_max([tf.constant(1.0, dtype=tmp.dtype), tmp])

        # Extract the wh offset target weight
        # wh_offset_target_weight[batch_id, :, cty_int, ctx_int] = 1
        # => 1 anywhere there is a target offset and wh
        tmp = tf.equal(tf.constant(0, dtype=y_true.dtype), target_offset)
        tmp = tf.logical_not(tmp)
        wh_offset_target_weight = tf.cast(tmp, dtype=tf.float32)
        return target_xy, target_wh, target_offset, avg_factor, wh_offset_target_weight

    def heatmap_loss(self, y_true, y_pred, avg_factor):
        """Implements `Gaussian Focal loss <https://arxiv.org/abs/1708.02002>`_
        for targets in gaussian distribution.

        Original source: mmdetection/losses/gaussian_focal_loss

        Args:
            y_true (tf.Tensor): tensor of true labels.
            y_pred (tf.Tensor): tensor of predicted labels.
            avg_factor (tf.Tensor): average factor.

        Returns:
            tf.Tensor: Heatmap loss
        """
        # Compute the loss
        pos_weights = tf.cast(tf.equal(y_true, 1.0), dtype=tf.float32)
        neg_weights = tf.math.pow((1 - y_true), self._gamma)
        pos_loss = -tf.math.log(y_pred + self._eps) * \
            tf.math.pow((1 - y_pred), self._alpha) * pos_weights
        neg_loss = -tf.math.log(1 - y_pred + self._eps) * \
            tf.math.pow(y_pred, self._alpha) * neg_weights
        loss = pos_loss + neg_loss
        # Compute the average across the matrix
        loss = tf.reduce_sum(loss) / avg_factor
        return loss

    def l1_loss(self, y_true, y_pred, avg_factor, weights=None):
        """L1 loss, used in location loss

        Args:
            y_true (tf.Tensor): tensor of true labels.
            y_pred (tf.Tensor): tensor of predicted labels.
            avg_factor (tf.Tensor): average factor.
            weights (tf.Tensor, optional): factor to multiply the loss. Defaults to None.

        Returns:
            tf.Tensor: L1 loss
        """
        difference = y_true - y_pred
        loss = tf.abs(difference)
        if weights is not None:
            loss *= weights
        loss = tf.reduce_sum(loss) / avg_factor
        return loss

    def __call__(self, y_true, y_pred_raw):
        # Get the avg factor and wh / offset weights
        (target_xy, target_wh, target_offset, avg_factor,
            wh_offset_target_weight) = self._get_targets(y_true)
        # Extract the 3 // branches + apply sigmoid
        y_pred_xy, y_pred_wh, y_pred_offset = self._transform_netout(y_pred_raw)

        # Heatmap loss
        center_heatmap_loss = self.heatmap_loss(target_xy, y_pred_xy, avg_factor)
        center_heatmap_loss *= self.heatmap_loss_weight

        # Wh loss
        wh_loss = self.l1_loss(target_wh, y_pred_wh, avg_factor * 2, wh_offset_target_weight)
        wh_loss *= self.wh_loss_weight

        # Offset loss
        offset_loss = self.l1_loss(target_offset, y_pred_offset, avg_factor * 2,
                                   wh_offset_target_weight)
        offset_loss *= self.offset_loss_weight

        loss = center_heatmap_loss + wh_loss + offset_loss
        return loss



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
AkidaNet18 model definition for ImageNet classification.

AkidaNet18 architecture is inspired both from AkidaNet and ResNet18: same depth and dimensions than
ResNet18 but without skip connections and using SeparableConvolution layers.

"""
from keras import Model, regularizers
from keras.layers import Input, Rescaling

from cnn2snn import quantize, load_quantized_model

from .imagenet_utils import obtain_input_shape
from ..layer_blocks import conv_block, separable_conv_block, dense_block
from ..utils import fetch_file


BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/AkidaV1/akidanet18/'


def akidanet18_imagenet(input_shape=None,
                        include_top=True,
                        pooling=None,
                        classes=1000,
                        depths=(4, 4, 4, 4),
                        dimensions=(64, 128, 256, 512),
                        weight_quantization=0,
                        activ_quantization=0,
                        input_weight_quantization=None,
                        input_scaling=(128, -1)):
    """Instantiates the AkidaNet18 architecture.

    Args:
        input_shape (tuple, optional): shape tuple. Defaults to None.
        include_top (bool, optional): whether to include the fully-connected
            layer at the top of the model. Defaults to True.
        pooling (str, optional): optional pooling mode for feature extraction
            when `include_top` is `False`.
            Defaults to None.

            * `None` means that the output of the model will be the 4D tensor
              output of the last convolutional block.
            * `avg` means that global average pooling will be applied to the
              output of the last convolutional block, and thus the output of the
              model will be a 2D tensor.
        classes (int, optional): optional number of classes to classify images
            into, only to be specified if `include_top` is `True`. Defaults to 1000.
        depth (tuple, optional): number of layers in each stages of the model. The length of the
            tuple defines the number of stages. Defaults to (4, 4, 4, 4).
        dimensions (tuple, optional): number of filters in each stage on the model. The length of
            the tuple must be equal to the length of the `depth` tuple. Defaults to
            (64, 128, 256, 512).
        weight_quantization (int, optional): sets all weights in the model to have
            a particular quantization bitwidth except for the weights in the
            first layer.
            Defaults to 0.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        activ_quantization (int, optional): sets all activations in the model to have a
            particular activation quantization bitwidth.
            Defaults to 0.

            * '0' implements floating point 32-bit activations.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_weight_quantization (int, optional): sets weight quantization in the first layer.
            Defaults to weight_quantization value.

            * '0' implements floating point 32-bit weights.
            * '2' through '8' implements n-bit weights where n is from 2-8 bits.
        input_scaling (tuple, optional): scale factor and offset to apply to
            inputs. Defaults to (128, -1). Note that following Akida convention,
            the scale factor is an integer used as a divider.

    Returns:
        keras.Model: a Keras model for AkidaNet/ImageNet.

    Raises:
        ValueError: in case of invalid input shape or mismatching `depth` and `dimensions`.
    """
    # Check if overrides have been provided and override
    if input_weight_quantization is None:
        input_weight_quantization = weight_quantization

    # Sanity checks
    stages = len(depths)
    if len(dimensions) != stages:
        raise ValueError(f"'depth' and 'dimensions' must be of the same length, received: {depths} "
                         f"and {dimensions}.")

    # Define weight regularization, will apply to the convolutional layers and
    # to all pointwise weights of separable convolutional layers.
    weight_regularizer = regularizers.l2(4e-5)

    # Determine proper input shape and default size.
    if input_shape is None:
        default_size = 224
    else:
        rows = input_shape[0]
        cols = input_shape[1]

        if rows == cols and rows in [128, 160, 192, 224]:
            default_size = rows
        else:
            default_size = 224

    input_shape = obtain_input_shape(input_shape,
                                     default_size=default_size,
                                     min_size=32,
                                     include_top=include_top)

    img_input = Input(shape=input_shape, name="input")

    if input_scaling is None:
        x = img_input
    else:
        scale, offset = input_scaling
        x = Rescaling(1. / scale, offset, name="rescaling")(img_input)

    # ConvNext stem layer: 4x4 kernel with stride 4
    x = conv_block(x,
                   filters=int(dimensions[0]),
                   name='convnext_stem',
                   kernel_size=(4, 4),
                   padding='same',
                   use_bias=False,
                   strides=4,
                   add_batchnorm=True,
                   add_activation=True,
                   kernel_regularizer=weight_regularizer)

    # Define the stages
    for stage in range(stages):
        # Like for AkidaNet, early layers (first 2 stages) are defined as standard Convolutional and
        # next layers are SeparableConvolutional layers
        if stage < 2:
            current_block = conv_block
            kwarg = {"kernel_regularizer": weight_regularizer}
        else:
            current_block = separable_conv_block
            kwarg = {"pointwise_regularizer": weight_regularizer}

        strides = 2 if stage > 0 else 1
        for i in range(depths[stage]):
            # First layer in stage comes with strides 2 except in first stage where strides is
            # handled by the previous stem
            strides = 2 if i == 0 and stage > 0 else 1

            # Handle final pooling in last layer of last stage
            if stage == stages - 1 and i == depths[stage] - 1:
                pool = 'global_avg' if include_top or pooling == 'avg' else None
            else:
                pool = None

            x = current_block(x,
                              filters=int(dimensions[stage]),
                              name=f'stage_{stage}/conv_{i}',
                              kernel_size=(3, 3),
                              strides=strides,
                              padding='same',
                              use_bias=False,
                              pooling=pool,
                              add_batchnorm=True,
                              add_activation=True,
                              **kwarg)

    # Classification layer
    if include_top:
        x = dense_block(x,
                        classes,
                        add_batchnorm=False,
                        add_activation=False,
                        kernel_initializer="he_normal",
                        name='classifier',
                        kernel_regularizer=weight_regularizer)

    # Create model
    model = Model(img_input, x, name='akidanet18_%s_%s' % (input_shape[0], classes))

    if ((weight_quantization != 0) or (activ_quantization != 0) or
            (input_weight_quantization != 0)):
        return quantize(model, weight_quantization, activ_quantization,
                        input_weight_quantization)
    return model


def akidanet18_imagenet_pretrained():
    """
    Helper method to retrieve an `akidanet18_imagenet` model that was trained on ImageNet dataset.

    Returns:
        keras.Model: a Keras Model instance.

    """
    model_name = 'akidanet18_imagenet_224_iq8_wq4_aq4.h5'
    file_hash = 'd3a0a3f37b5d3a5998981705e772b40670fdab0dd68e5894dc8af717bbaae136'
    model_path = fetch_file(BASE_WEIGHT_PATH + model_name,
                            fname=model_name,
                            file_hash=file_hash,
                            cache_subdir='models')
    return load_quantized_model(model_path)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Tools to unfuse SeparableConv2D layers from a model.
"""

__all__ = ["unfuse_sepconv2d"]

from copy import deepcopy

from keras.models import Sequential
from keras.layers import SeparableConv2D
from quantizeml.models.utils import apply_weights_to_model
from quantizeml.models.transforms.transforms_utils import (
    get_layer_index, get_layers_by_type, update_inbound)


def _find_sepconv_layers(model):
    """ Retrieves SeparableConv2D layers.

    Args:
        model (keras.Model): a model

    Returns:
        dict: map between a SeparableConv2D and the layer that follows
    """
    map_sepconv_next = {}

    # Get all SeparableConv2D layers present in the model
    sep_conv_layers = get_layers_by_type(model, SeparableConv2D)

    for sep_conv in sep_conv_layers:
        # Limit support to single inbound/outbound
        outbounds = sep_conv.outbound_nodes
        if len(sep_conv.inbound_nodes) != 1 or len(outbounds) != 1:
            continue

        following_layer = outbounds[0].layer
        # At this point the SeparableConv2D is a valid candidate
        map_sepconv_next[sep_conv] = following_layer

    # Check if the last model layer is a SeparableConv2D if it's the case add it
    # to dict but with None as a special value (to indicate that it has no outbound layer)
    last_layer = model.layers[-1]
    if isinstance(last_layer, SeparableConv2D):
        map_sepconv_next[last_layer] = None
    return map_sepconv_next


def _get_depthwise_conf(sep_layer):
    """ Helper function to unfuse a SeparableConv2D layer. This function returns the
        corresponding DepthwiseConv2D layer configuration and its generated name.

    Args:
        sep_layer (keras.Layer): a keras SeparableConv2D layer

    Returns:
        dict: the DepthwiseConv2D layer configuration
    """
    assert isinstance(sep_layer, SeparableConv2D), ("_get_depthwise_conf accepts only "
                                                    "SeparableConv2D layers. "
                                                    f"Got {type(sep_layer)}")
    dw_name = "dw_" + sep_layer.name
    sep_layer_config = sep_layer.get_config()
    depthwise_initializer = sep_layer_config.get("depthwise_initializer",
                                                 {'class_name': 'GlorotUniform',
                                                  'config': {'seed': None}})
    depthwise_regularizer = sep_layer_config.get("depthwise_regularizer", None)
    depthwise_constraint = sep_layer_config.get("depthwise_constraint", None)

    dw_layer_conf = {'class_name': 'DepthwiseConv2D',
                     'config': {'name': dw_name, 'trainable': sep_layer.trainable,
                                'dtype': sep_layer.dtype,
                                'kernel_size': sep_layer.kernel_size,
                                'strides': sep_layer.strides,
                                'padding': sep_layer.padding, 'data_format': sep_layer.data_format,
                                'dilation_rate': sep_layer.dilation_rate,
                                'groups': sep_layer.groups, 'activation': 'linear',
                                'use_bias': False,
                                'bias_initializer': {'class_name': 'Zeros', 'config': {}},
                                'bias_regularizer': None,
                                'bias_constraint': None,
                                'activity_regularizer': None,
                                'depth_multiplier': sep_layer.depth_multiplier,
                                'depthwise_initializer': depthwise_initializer,
                                'depthwise_regularizer': depthwise_regularizer,
                                'depthwise_constraint': depthwise_constraint}}

    return dw_layer_conf


def _get_pointwise_conf(sep_layer):
    """ Helper function to unfuse a SeparableConv2D layer. This function returns the
        corresponding (pointwise)Conv2D layer configuration and its generated name.

    Args:
        sep_layer (keras.Layer): a keras SeparableConv2D layer

    Returns:
        dict: the (pointwise)Conv2D layer configuration
    """
    assert isinstance(sep_layer, SeparableConv2D), ("_get_pointwise_conf accepts only "
                                                    "SeparableConv2D layers. "
                                                    f"Got {type(sep_layer)}")
    pw_name = "pw_" + sep_layer.name
    sep_layer_config = sep_layer.get_config()
    pointwise_initializer = sep_layer_config.get("pointwise_initializer",
                                                 {'class_name': 'GlorotUniform',
                                                  'config': {'seed': None}})
    pointwise_regularizer = sep_layer_config.get("pointwise_regularizer", None)
    pointwise_constraint = sep_layer_config.get("pointwise_constraint", None)
    bias_initializer = sep_layer_config.get("bias_initializer",
                                            {'class_name': 'Zeros', 'config': {}})
    bias_regularizer = sep_layer_config.get("bias_regularizer", None)
    bias_constraint = sep_layer_config.get("bias_constraint", None)
    activation = sep_layer_config.get("activation", "linear")
    activity_regularizer = sep_layer_config.get("activity_regularizer", None)

    pw_layer_conf = {'class_name': 'Conv2D',
                     'config': {'name': pw_name, 'trainable': sep_layer.trainable,
                                'dtype': sep_layer.dtype, 'filters': sep_layer.filters,
                                'kernel_size': (1, 1), 'strides': (1, 1),
                                'padding': 'valid', 'data_format': sep_layer.data_format,
                                'dilation_rate': (1, 1), 'groups': sep_layer.groups,
                                'activation': activation, 'use_bias': sep_layer.use_bias,
                                'kernel_initializer': pointwise_initializer,
                                'bias_initializer': bias_initializer,
                                'kernel_regularizer': pointwise_regularizer,
                                'bias_regularizer': bias_regularizer,
                                'activity_regularizer': activity_regularizer,
                                'kernel_constraint': pointwise_constraint,
                                'bias_constraint': bias_constraint}}

    return pw_layer_conf


def _get_unfused_sepconv_model(model, map_sepconv_layers):
    """Edits the model configuration to unfuse SeparableConv2D layers and rebuilds a model.
    Returns also a dict mapping the new unfused layers variable names with their
    corresponding values (i.e from the main SeparableConv2D layer).

    Args:
        model (keras.Model): a model
        map_sepconv_layers (dict): map between a SeparableConv2D and the layer that follows (None
            if the SeparableConv2D layer is the last model layer)

    Returns:
        tuple (keras.Model, dict): (an updated model with unfused SeparableConv2D layers,
        new layers variables)
    """
    # get_config documentation mentions that a copy should be made when planning to modify the
    # config
    config = deepcopy(model.get_config())
    layers = config['layers']

    # Create an empty dict that will be populated with the new layers variables names
    # and their corresponding values
    sep_layers_vars_dict = {}

    for sep_conv, next_layer in map_sepconv_layers.items():
        # Get the index of the SeparableConv2D layer
        sep_conv_index = get_layer_index(layers, sep_conv.name)

        # Get the original SeparableConv2D layer inbound informations if available
        sep_conv_inbounds_conf = layers[sep_conv_index].get('inbound_nodes', None)

        # Get DepthwiseConv2D conf from the main SeparableConv2D layer
        dw_conf = _get_depthwise_conf(sep_conv)
        # Replace the SeparableConv2D by a DepthwiseConv2D
        layers[sep_conv_index] = dw_conf

        # And insert after it a Conv2D layer (i.e pointwise conv2d)
        pw_conf = _get_pointwise_conf(sep_conv)
        layers.insert(sep_conv_index + 1, pw_conf)

        # Get new layers names
        dw_name = dw_conf['config']['name']
        pw_name = pw_conf['config']['name']
        # Get new layers var names
        dw_kernel_name = dw_name+"/depthwise_kernel:0"
        pw_kernel_name = pw_name+"/kernel:0"
        pw_bias_name = pw_name+"/bias:0"
        # Map new vars with the corresponding ones from the SeparableConv2D layer
        sep_layers_vars_dict[dw_kernel_name] = sep_conv.depthwise_kernel
        sep_layers_vars_dict[pw_kernel_name] = sep_conv.pointwise_kernel
        sep_layers_vars_dict[pw_bias_name] = sep_conv.bias

        # For sequential model, the changes stop here: the SeparableConv2D layers will simply be
        # replaced by an equivalent DepthwiseConv2D + (pointwise)Conv2D layers. For other models,
        # the layers inbounds/outbounds must be rebuilt.
        if isinstance(model, Sequential):
            continue

        # Retrieve the new DepthwiseConv2D index
        dw_layer_index = get_layer_index(layers, dw_name)
        layers[dw_layer_index]['name'] = dw_name
        # Set the new DepthwiseConv2D layer inbounds (those are the main SeparableConv2D inbounds)
        layers[dw_layer_index]['inbound_nodes'] = sep_conv_inbounds_conf

        # Retrieve the new (pointwise)Conv2D index (dw_layer_index + 1)
        pw_layer_index = dw_layer_index + 1
        layers[pw_layer_index]['name'] = pw_name
        # Set the new (pointwise)Conv2D layer inbounds (its the new DepthwiseConv2D layer)
        # tfmot code: 'inbound_nodes' is a nested list where first element is the inbound layername,
        # e.g: [[['conv1', 0, 0, {} ]]]
        layers[pw_layer_index]['inbound_nodes'] = [[[dw_name, 0, 0, {}]]]

        if next_layer:
            # Retrieve the next layer index
            next_layer_index = get_layer_index(layers, next_layer.name)
            # Update the next layer inbound layer (i.e the main SeparableConv2D layer) by the
            # new (pointwise)Conv2D
            update_inbound(layers[next_layer_index], sep_conv.name, pw_name)
        else:
            config['output_layers'] = [[pw_name, 0, 0]]

    # Reconstruct model from the config, using the cloned layers
    return model.from_config(config), sep_layers_vars_dict


def unfuse_sepconv2d(model):
    """ Unfuse the SeparableConv2D layers of a model by replacing them with an equivalent
    DepthwiseConv2D + (pointwise)Conv2D layers.

    Args:
        model (keras.Model): the model to update

    Returns:
        keras.Model: the original model or a new model with unfused SeparableConv2D layers
    """
    # Find SeparableConv2D
    map_sepconv_layers = _find_sepconv_layers(model)

    # When there are no valid candidates, return the original model
    if not map_sepconv_layers:
        return model

    # Rebuild a model with unfused SeparableConv2D by editing the configuration
    updated_model, sep_layers_vars_dict = _get_unfused_sepconv_model(model, map_sepconv_layers)

    # Load original weights
    variables_dict = {}

    for layer in model.layers:
        if not isinstance(layer, SeparableConv2D):
            for var in layer.variables:
                variables_dict[var.name] = var

    variables_dict.update(sep_layers_vars_dict)
    apply_weights_to_model(updated_model, variables_dict, False)

    return updated_model



#!/usr/bin/env python
# coding: utf-8
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for ImageNet models.
"""

import os
import argparse

import numpy as np
import tensorflow as tf
try:
    import tensorflow_datasets as tfds
except ImportError:
    tfds = None

import keras
from keras.callbacks import LearningRateScheduler
from keras import Sequential
from keras.optimizers import SGD, Adam
from tensorflow_addons.optimizers import LAMB
from keras.layers import Input
from keras.models import clone_model

import akida
from cnn2snn import convert
from quantizeml.layers import AddPositionEmbs, ClassToken

from .preprocessing import preprocess_image, DATA_AUGMENTATION
from ..training import (get_training_parser, calibrate_model, freeze_model_before,
                        print_history_stats, RestoreBest, save_model)
from ..calibrable import calibration_required
from ..transformers.model_vit import apply_embedding_weights
from ..distiller import DeitDistiller
from ..param_scheduler import get_cosine_lr_scheduler
from ..utils import get_tensorboard_callback, load_model


def get_imagenet_dataset(data_path, training, image_size, batch_size, data_aug=True, one_hot=False):
    """ Loads ImageNet 2012 dataset and builds a tf.dataset out of it.

    Args:
        data_path (str): path to the folder containing ImageNet tar files
        training (bool): True to retrieve training data, False for validation
        image_size (int): desired image size
        batch_size (int): the batch size
        data_aug (bool, optional): True to apply data augmentation (only train). Defaults to True.
        one_hot (bool, optional): whether to one hot labels or not. Defaults to False.

    Returns:
        tf.dataset, int: the requested dataset (train or validation) and the
        corresponding steps
    """
    assert tfds is not None, "To load imagenet dataset, tensorflow-datasets module must\
        be installed."
    # Build the dataset
    write_dir = os.path.join(data_path, 'tfds')

    download_and_prepare_kwargs = {
        'download_dir': os.path.join(write_dir, 'downloaded'),
        'download_config': tfds.download.DownloadConfig(manual_dir=data_path)
    }

    split = 'train' if training else 'validation'

    dataset, infos = tfds.load(
        'imagenet2012',
        data_dir=os.path.join(write_dir, 'data'),
        split=split,
        shuffle_files=training,
        download=True,
        as_supervised=True,
        download_and_prepare_kwargs=download_and_prepare_kwargs,
        with_info=True)

    if training:
        dataset = dataset.shuffle(10000, reshuffle_each_iteration=True).repeat()

    data_aug = DATA_AUGMENTATION if data_aug else None
    dataset = dataset.map(lambda image, label: (preprocess_image(
        image, image_size, training, data_aug), label))

    # One hot encode labels if requested
    if one_hot:
        num_classes = infos.features["label"].num_classes
        dataset = dataset.map(lambda image, label: (image, tf.one_hot(label, num_classes)))

    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(1)

    # The following will silence a Tensorflow warning on auto shard policy
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    dataset = dataset.with_options(options)

    return dataset, infos.splits[split].num_examples / batch_size


def compile_model(model, optimizer="SGD", distiller_type="none"):
    """ Compiles the model.

    Args:
        model (keras.Model): the model to compile
        optimizer (str, optional): the optimizer to use. Defaults to "SGD".
        distiller_type (str, optional): string to select the loss in distillation.
            Only used when input model is of ``DeitDistiller`` type. Defaults to 'none'.

    Returns:
        bool: True if labels should be one-hot encoded, False if not.
    """

    def _get_optim(optim_str):
        optim_str_low = optim_str.lower()
        if optim_str_low == "sgd":
            return SGD(momentum=0.9)
        elif optim_str_low == "adam":
            return Adam(epsilon=1e-8)
        elif optim_str_low == "lamb":
            return LAMB(epsilon=1e-8, weight_decay_rate=2e-2)
        else:
            raise ValueError(f"Unknown optimizer {optim_str}. "
                             "Please choose one of these options: {SGD, ADAM, LAMB}")

    if isinstance(model, DeitDistiller):
        model.compile(optimizer=_get_optim(optimizer),
                      metrics=['accuracy', 'top_k_categorical_accuracy'],
                      student_loss_fn=keras.losses.CategoricalCrossentropy(from_logits=True),
                      distiller_type=distiller_type)
        return True

    model.compile(optimizer=_get_optim(optimizer),
                  loss=keras.losses.CategoricalCrossentropy(from_logits=True),
                  metrics=['accuracy', 'top_k_categorical_accuracy'])
    return True


def evaluate(model, val_dataset, batch_size, num_samples, val_steps):
    """ Evaluates model performances.

    Args:
        model (keras.Model or akida.Model): the model to compile evaluate
        val_dataset (tf.dataset): validation dataset
        batch_size (int): the batch size
        num_samples (int): number of samples to use for Akida
        val_steps (int): validation steps
    """
    if isinstance(model, akida.Model):
        correct_preds = 0
        cur_samples = 0
        total_samples = val_steps * batch_size
        if num_samples <= 0:
            num_samples = total_samples
        else:
            num_samples = min(num_samples, total_samples)
        it = val_dataset.as_numpy_iterator()

        print(f"Processing {num_samples} samples.")
        if num_samples > batch_size:
            n_batches = num_samples // batch_size
            if n_batches > 5:
                log_samples = (n_batches // 5) * batch_size
            else:
                log_samples = batch_size
            print(f"Logging every {log_samples} samples.")
        else:
            log_samples = num_samples
        while cur_samples < num_samples:
            x, y = next(it)
            y_ak = model.predict_classes(x.astype(np.uint8))
            correct_preds += np.sum(y_ak == np.argmax(y, -1))
            cur_samples += y_ak.shape[0]
            if cur_samples % log_samples == 0 and cur_samples < num_samples:
                # Log current accuracy
                accuracy = correct_preds / cur_samples
                print(f"Accuracy after {cur_samples}: {accuracy}")
        accuracy = correct_preds / cur_samples
        print(f"Accuracy after {cur_samples}: {accuracy}")
    else:
        history = model.evaluate(val_dataset, steps=val_steps)
        print(history)


def rescale_legacy(base_model, input_size):
    """ Rescales the model by changing its input size.

    Args:
        base_model (keras.Model): the model to rescale
        input_size (int): desired model input size

    Returns:
        keras.Model: the rescaled model
    """
    # Create the desired input
    input_shape = (input_size, input_size, base_model.input.shape[-1])
    new_input = Input(input_shape)

    # Workaround to force the input shape update that is not working for
    # functional models: the input_tensors parameter is ignored as described in
    # https://github.com/tensorflow/tensorflow/issues/40617.
    if not isinstance(base_model, Sequential):
        base_model.layers[0]._batch_input_shape = (None, input_size, input_size,
                                                   base_model.input.shape[-1])
        new_input = None

    # Clone the model and replace input layer
    clone = clone_model(base_model, input_tensors=new_input)
    clone.set_weights(base_model.get_weights())
    return clone


def rescale_vit(base_model, input_size):
    """ Rescales the model by changing its input size in three steps:
        1. Change the input layer, given the new input size values
        2. Reconstruct model from config
        3. Update weights, taking into account interpolation on PosEmbed layer

    Args:
        base_model (keras.Model): the model to rescale
        input_size (int): desired model input size

    Returns:
        keras.Model: the rescaled model
    """
    # 1. Create the desired input layer:
    input_shape = (None, input_size, input_size, base_model.input.shape[-1])
    num_tokens = sum(isinstance(ly, ClassToken) for ly in base_model.layers)

    # 2. Clone the model by modification of dict_config:
    # In models based on vision transformers, PositionEmbeding change its amount of parameters
    # given an input size. This is why we need to reconstruct the model.
    x_patch, y_patch = None, None
    clone_config = base_model.get_config()

    for layer_config in clone_config["layers"]:
        # 2.1. Change input size
        if layer_config["class_name"] == "InputLayer":
            layer_config["config"]["batch_input_shape"] = input_shape
        # 2.2. Recover total of patches in both directions
        if "Conv2D" in layer_config["class_name"]:
            x_patch = input_size // layer_config["config"]["kernel_size"][0]
            y_patch = input_size // layer_config["config"]["kernel_size"][1]
        # 2.3. Change values in reshape process
        elif layer_config["class_name"] == "Reshape":
            layer_config["config"]["target_shape"] = list(layer_config["config"]["target_shape"])
            layer_config["config"]["target_shape"][0] = x_patch * y_patch

    # 2.5. Recompile model with new configuration
    clone = base_model.from_config(clone_config)

    # 3. Update weights:
    # Get weights from based model, and tranfer them into clone model,
    for base_layer, clone_layer in zip(base_model.layers, clone.layers):
        base_weights = base_layer.get_weights()
        if isinstance(base_layer, AddPositionEmbs):
            apply_embedding_weights(clone_layer, base_weights, (x_patch, y_patch), num_tokens)
        else:
            clone_layer.set_weights(base_weights)
    return clone


def rescale(base_model, input_size):
    """ Rescale the model by architecture (if there is a vision transformer model or not)

    Args:
        base_model (keras.Model): the model to rescale
        input_size (int): desired model input size

    Returns:
        keras.Model: the rescaled model
    """
    is_vit = any(isinstance(layer, AddPositionEmbs) for layer in base_model.layers)
    if is_vit:
        return rescale_vit(base_model, input_size)
    return rescale_legacy(base_model, input_size)


def train(model,
          train_dataset,
          train_steps,
          val_dataset,
          val_steps,
          out_dir,
          num_epochs,
          tune=False,
          learning_rate=1e-1,
          initial_epoch=0,
          lr_policy='exp_decay'):
    """ Trains the model

    Args:
        model (keras.Model): the model to train
        train_dataset (tf.dataset): training dataset
        train_steps (int): train steps
        val_dataset (tf.dataset): validation dataset
        val_steps (int): validation steps
        out_dir (str): parent directory for logs folder
        num_epochs (int): the number of epochs
        tune (bool, optional): enable tuning (lower learning rate). Defaults to
          False.
        learning_rate (float, optional): the learning rate. Defaults to 1e-1.
        initial_epoch (int, optional): epoch at which to start training.
          Defaults to 0.
        lr_policy (str, optional): defines the learning rate policy to adopt. Values in
            ['exp_decay', 'cosine_decay', 'cosine_sched'] for exponential decay, cosine decay and
            cosine oscillation respectively. Defaults to 'exp_decay'.
    """
    # 1. Check model validity
    if calibration_required(model):
        str_action = "tune" if tune else "train"
        raise ValueError(f"{model.name} model must be calibrated before a {str_action}.")

    # 2. Define training callbacks
    callbacks = []

    # 2.1 Learning rate scheduler
    if lr_policy == 'exp_decay':
        LR_START = learning_rate
        LR_END = 1e-4
        # number of epochs you first keep the learning rate constant
        LR_EPOCH_CONSTANT = 10
        # Modify default values for fine-tuning
        if tune:
            LR_START = 1e-4
            LR_END = 1e-8
            LR_EPOCH_CONSTANT = 2

        if LR_EPOCH_CONSTANT >= num_epochs:
            lr_decay = LR_END / LR_START
        else:
            lr_decay = (LR_END / LR_START)**(1. / (num_epochs - LR_EPOCH_CONSTANT))

        # This function keeps the learning rate at LR_START for the first N epochs
        # and decreases it exponentially after that.
        def agg_lr_scheduler(epoch):
            if epoch < LR_EPOCH_CONSTANT:
                return LR_START
            return LR_START * lr_decay**(epoch - (LR_EPOCH_CONSTANT - 1))

        lr_scheduler = LearningRateScheduler(agg_lr_scheduler)
    elif lr_policy == 'cosine_decay':
        LR_START = learning_rate
        if tune:
            LR_EPOCH_CONSTANT = 2  # number of epochs you first keep the learning rate constant
            LR_MIN = 1e-8
        else:
            LR_EPOCH_CONSTANT = 5  # number of epochs you first keep the learning rate constant
            LR_MIN = 1e-6
        # Make sure to start with a learning rate higher than LR_MIN
        LR_MIN = LR_MIN if LR_START > LR_MIN else 0.1 * LR_START

        # If the number of epochs is too small, dont use WARMUP_LR
        if num_epochs < LR_EPOCH_CONSTANT:
            LR_EPOCH_CONSTANT = 0

        # This function keeps the learning rate at LR_START for the first N epochs and decreases it
        # following
        # https://www.tensorflow.org/api_docs/python/tf/keras/optimizers/schedules/CosineDecay
        def cos_lr_scheduler(epoch):
            if epoch < LR_EPOCH_CONSTANT:
                return LR_START
            step = min(epoch, num_epochs) - LR_EPOCH_CONSTANT
            cosine_decay = 0.5 * (1 + np.cos(np.pi * step / (num_epochs - LR_EPOCH_CONSTANT)))
            decayed = (1 - LR_MIN) * cosine_decay + LR_MIN
            return LR_START * decayed
        lr_scheduler = LearningRateScheduler(cos_lr_scheduler)
    elif lr_policy == 'cosine_sched':
        lr_scheduler = get_cosine_lr_scheduler(learning_rate, num_epochs * train_steps)
    else:
        raise ValueError(f"Unsupported learning rate policy '{lr_policy}'.")
    callbacks.append(lr_scheduler)

    # 2.2 Model checkpoints (save best model and retrieve it when training is complete)
    restore_model = RestoreBest(model)
    callbacks.append(restore_model)

    # 2.3 Tensorboard logs
    tensorboard = get_tensorboard_callback(out_dir, prefix='imagenet')
    callbacks.append(tensorboard)

    # 3. Train
    history = model.fit(train_dataset,
                        steps_per_epoch=train_steps,
                        epochs=num_epochs,
                        callbacks=callbacks,
                        validation_data=val_dataset,
                        validation_steps=val_steps,
                        initial_epoch=initial_epoch)
    print_history_stats(history)


def main():
    """ Entry point for script and CLI usage.
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument(
        "-d",
        "--data_dir",
        type=str,
        default='/hdd/datasets/imagenet/',
        help="The directory containing the ImageNet data.")
    global_parser.add_argument("-o",
                               "--out_dir",
                               type=str,
                               default='./logs',
                               help="The output directory (logs, checkpoints).")

    parsers = get_training_parser(batch_size=128,
                                  calibrate=True,
                                  freeze_before=True,
                                  tune=True,
                                  global_parser=global_parser)

    train_parser = parsers[1]
    train_parser.add_argument("-lr",
                              "--learning_rate",
                              type=float,
                              default=1e-1,
                              help="Learning rate start value.")
    train_parser.add_argument("-ie",
                              "--initial_epoch",
                              type=int,
                              default=0,
                              help="Epoch at which to start training.")
    train_parser.add_argument("--optim", type=str, default="SGD",
                              help="Optimizer to use. Defaults to %(default)s.")
    train_parser.add_argument("--data_aug", action='store_true', help="Enables custom DA.")
    train_parser.add_argument("--lr_policy",
                              default='exp_decay',
                              choices=['exp_decay', 'cosine_decay', 'cosine_sched'],
                              help="Defines the learning rate scheduling. Values in "
                                   "['exp_decay', 'cosine_decay', 'cosine_sched'] which corresponds"
                                   " to exponential decay, cosine decay and cosine oscillation "
                                   "respectively.")
    train_parser.add_argument("--teacher", type=str, default=None,
                              help="Teacher model use to train the model through Knowledge"
                              "Distillation. The input model output an Add layer. "
                              "Defaults to %(default)s.")
    train_parser.add_argument("-dt", "--distiller_type", type=str, default="soft",
                              help="Define the distillation loss type. Defaults to %(default)s.")
    train_parser.add_argument("-a", "--alpha", type=float, default=0.5,
                              help="Value for distiller losses weighting. Defaults to %(default)s.")

    tune_parser = parsers[2]
    tune_parser.add_argument("-ie",
                             "--initial_epoch",
                             type=int,
                             default=0,
                             help="Epoch at which to start training.")
    tune_parser.add_argument("-lr",
                             "--learning_rate",
                             type=float,
                             default=6e-5,
                             help="Learning rate start value.")
    tune_parser.add_argument("--data_aug",
                             action='store_true',
                             help="Enables custom DA.")
    tune_parser.add_argument("--optim",
                             type=str,
                             default="SGD",
                             help="Optimizer to use. Defaults to %(default)s.")
    tune_parser.add_argument("--lr_policy",
                             default='exp_decay',
                             choices=['exp_decay', 'cosine_decay', 'cosine_sched'],
                             help="Defines the learning rate scheduling. Values in "
                                  "['exp_decay', 'cosine_decay', 'cosine_sched'] which corresponds "
                                  "to exponential decay, cosine decay and cosine oscillation "
                                  "respectively.")
    tune_parser.add_argument("--teacher", type=str, default=None,
                             help="Teacher model use to train the model through Knowledge"
                             "Distillation. The input model output an Add layer. "
                             "Defaults to %(default)s.")
    tune_parser.add_argument("-dt", "--distiller_type", type=str, default="soft",
                             help="Define the distillation loss type. Defaults to %(default)s.")
    tune_parser.add_argument("-a", "--alpha", type=float, default=0.5,
                             help="Value for distiller losses weighting. Defaults to %(default)s.")

    eval_parser = parsers[3]
    eval_parser.add_argument("-ns",
                             "--num_samples",
                             type=int,
                             default=-1,
                             help="Number of samples to use (for Akida)")

    subparsers = parsers[-1]
    rescale_parser = subparsers.add_parser("rescale",
                                           help="Rescale a model.",
                                           parents=[global_parser])
    rescale_parser.add_argument("-i",
                                "--input_size",
                                type=int,
                                required=True,
                                help="The square input image size")
    rescale_parser.add_argument("-s",
                                "--savemodel",
                                type=str,
                                default=None,
                                help="Save model with the specified name")

    args = parsers[0].parse_args()

    # Load the source model
    model = load_model(args.model)

    # Try to load the teacher model, used after to train with Knowledge Distillation
    # Hyperparameters takes from https://arxiv.org/pdf/2012.12877.pdf
    if getattr(args, 'teacher', False):
        if args.distiller_type not in ['none', 'soft', 'hard']:
            raise ValueError("Distiller type must be one of ['none', 'soft', 'hard']")
        teacher = load_model(args.teacher)
        train_model = DeitDistiller(model, teacher, alpha=args.alpha, temperature=3.0)
    else:
        train_model = model

    # Freeze the model
    if "freeze_before" in args:
        freeze_model_before(model, args.freeze_before)

    # Compile model
    one_hot = compile_model(train_model, optimizer=getattr(args, "optim", "SGD"),
                            distiller_type=getattr(args, "distiller_type", "none"))

    # Load validation data
    if args.action != 'rescale':
        im_size = model.input_shape[1]
        val_ds, val_steps = get_imagenet_dataset(args.data_dir, False, im_size,
                                                 args.batch_size, data_aug=False, one_hot=one_hot)

    # Disable QuantizeML assertions
    os.environ["ASSERT_ENABLED"] = "0"

    # Train model
    if args.action in ['train', 'tune']:
        # Load training data
        train_ds, train_steps = get_imagenet_dataset(args.data_dir, True,
                                                     im_size, args.batch_size,
                                                     args.data_aug, one_hot=one_hot)
        tune = args.action == 'tune'

        learning_rate = args.learning_rate
        if args.lr_policy == 'cosine_decay':
            # Tune learning rate following https://arxiv.org/pdf/2012.12877.pdf
            learning_rate *= args.batch_size / 512

        train(train_model,
              train_ds,
              train_steps,
              val_ds,
              val_steps,
              args.out_dir,
              args.epochs,
              tune=tune,
              learning_rate=learning_rate,
              initial_epoch=args.initial_epoch,
              lr_policy=args.lr_policy)
        save_model(model, args.model, args.savemodel, args.action)

    elif args.action == 'eval':
        # Evaluate model accuracy
        if args.akida:
            model = convert(model)
        evaluate(model, val_ds, args.batch_size, args.num_samples, val_steps)

    elif args.action == 'rescale':
        # Rescale model
        m = rescale(model, args.input_size)
        save_model(m, args.model, args.savemodel, args.action)

    elif args.action == "calibrate":
        # Calibrate model
        # Load training data without data augmentation
        train_ds, train_steps = get_imagenet_dataset(args.data_dir, True,
                                                     im_size, args.batch_size,
                                                     False, one_hot=one_hot)
        calibrate_model(model, train_ds, args.num_samples, args.batch_size, args.epochs)
        save_model(model, args.model, args.savemodel, args.action)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
ViT model definition.
Inspired from https://github.com/faustomorales/vit-keras/blob/master/vit_keras/vit.py.
"""

import keras
import warnings
import numpy as np
import scipy as sp
import typing_extensions as tx

from quantizeml.layers import AddPositionEmbs, ClassToken, ExtractToken
from quantizeml.models import load_model

from ..layer_blocks import transformer_block, norm_to_layer
from ..utils import fetch_file

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/AkidaV2/vit/'


ConfigDict = tx.TypedDict(
    "ConfigDict",
    {
        "dropout": float,
        "mlp_dim": int,
        "num_heads": int,
        "num_blocks": int,
        "hidden_size": int,
    },
)

CONFIG_TI: ConfigDict = {
    "dropout": 0.0,
    "mlp_dim": 768,
    "num_heads": 3,
    "num_blocks": 12,
    "hidden_size": 192,
}

CONFIG_S: ConfigDict = {
    "dropout": 0.1,
    "mlp_dim": 1536,
    "num_heads": 6,
    "num_blocks": 12,
    "hidden_size": 384,
}

CONFIG_B: ConfigDict = {
    "dropout": 0.1,
    "mlp_dim": 3072,
    "num_heads": 12,
    "num_blocks": 12,
    "hidden_size": 768,
}

CONFIG_L: ConfigDict = {
    "dropout": 0.1,
    "mlp_dim": 4096,
    "num_heads": 16,
    "num_blocks": 24,
    "hidden_size": 1024,
}


def apply_embedding_weights(target_layer, source_weights, num_patches, num_tokens=1):
    """Apply embedding weights to a target layer.

    Args:
        target_layer (:obj:`keras.Layer`): The target layer to which weights will be applied.
        source_weights (list of :obj:`np.array`): The source weights.
        num_patches (int or tuple): Number of patches in xy spatial dimension, extracted from
            embedding layer.
        num_tokens (int, optional): Number of tokens. Defaults to 1.
    """
    assert isinstance(source_weights, list), "source_weights must be a list of numpy arrays"
    expected_shape = target_layer.weights[0].shape

    # number of patches constrains
    if isinstance(num_patches, int):
        num_patches = (num_patches, num_patches)
    assert len(num_patches) == 2, "num_patches must contain at most two values"
    assert all(x > 0 for x in num_patches), "every patch must be greater than 0"

    if expected_shape != source_weights[0].shape:
        token, grid = source_weights[0][0, :num_tokens], source_weights[0][0, num_tokens:]
        sin = int(np.sqrt(grid.shape[0]))
        sout_x = num_patches[0]
        sout_y = num_patches[1]
        warnings.warn(
            "Resizing position embeddings from " f"{sin}, {sin} to {sout_x}, {sout_y}",
            UserWarning,
        )
        zoom = (sout_y / sin, sout_x / sin, 1)
        grid = sp.ndimage.zoom(grid.reshape(sin, sin, -1), zoom, order=1).reshape(
            sout_x * sout_y, -1
        )
        new_weights = np.concatenate([token, grid], axis=0)[np.newaxis]
        source_weights = [new_weights] + source_weights[1:]
    target_layer.set_weights(source_weights)


def vit_imagenet(input_shape,
                 patch_size,
                 num_blocks,
                 hidden_size,
                 num_heads,
                 name,
                 mlp_dim,
                 classes=1000,
                 dropout=0.1,
                 include_top=True,
                 norm='LN',
                 last_norm='LN',
                 softmax='softmax',
                 act="GeLU"):
    """Build a ViT model.

    Args:
        input_shape (tuple): image shape tuple
        patch_size (int): the size of each patch (must fit evenly in image size)
        num_blocks (int): the number of transformer blocks to use.
        hidden_size (int): the number of filters to use
        num_heads (int): the number of transformer heads
        name (str): the model name
        mlp_dim (int): the number of dimensions for the MLP output in the transformers.
        classes (int, optional): number of classes to classify images into, only to be specified if
            `include_top` is True. Defaults to 1000.
        dropout (float, optional): fraction of the units to drop for dense layers. Defaults to 0.1.
        include_top (bool, optional): whether to include the final classifier head. If False,
            the output will correspond to that of the transformer. Defaults to True.
        norm (str, optional): string that values in ['LN', 'GN1', 'BN', 'LMN'] and that allows to
            choose from LayerNormalization, GroupNormalization(groups=1, ...), BatchNormalization
            or LayerMadNormalization layers respectively in the model. Defaults to 'LN'.
        last_norm (str, optional): string that values in ['LN', 'BN']
            and that allows to choose from LayerNormalization or
            BatchNormalization in the classifier network. Defaults to 'LN'.
        softmax (str, optional): string with values in ['softmax', 'softmax2']
            that allows to choose between softmax and softmax2 in MHA. Defaults
            to 'softmax'.
        act (str, optional): string that values in ['GeLU', 'ReLUx', 'swish'] and that allows to
            choose from GeLU, ReLUx or swish activation in MLP block. Defaults to 'GeLU'.
    """
    assert (input_shape[0] % patch_size == 0) and (
        input_shape[1] % patch_size == 0), "image size must be a multiple of patch_size"

    if last_norm not in ('LN', 'BN'):
        raise NotImplementedError("last_norm should be in ['LN', 'BN']"
                                  f"but received {norm}.")

    # Normalize image adding rescaling layer
    x = keras.layers.Input(shape=input_shape, name="input")
    y = keras.layers.Rescaling(1 / 127.5, -1, name="Rescale")(x)

    # Build model
    y = keras.layers.Conv2D(
        filters=hidden_size,
        kernel_size=patch_size,
        strides=patch_size,
        padding="valid",
        name="Embedding",
        kernel_initializer=keras.initializers.TruncatedNormal(stddev=0.02),
        bias_initializer="zeros",
    )(y)
    y = keras.layers.Reshape((y.shape[1] * y.shape[2], hidden_size))(y)
    y = ClassToken(name="ClassToken")(y)
    y = AddPositionEmbs(name="Transformer/PosEmbed")(y)
    for n in range(num_blocks):
        y, _ = transformer_block(
            y,
            num_heads=num_heads,
            hidden_size=hidden_size,
            mlp_dim=mlp_dim,
            dropout=dropout,
            name=f"Transformer/EncoderBlock_{n}",
            norm=norm,
            softmax=softmax,
            mlp_act=act,
        )

    # Include classification head
    if include_top:
        y = norm_to_layer(last_norm)(
            epsilon=1e-6, name="Transformer/EncoderNorm")(y)
        y = ExtractToken(token=0, name="ExtractToken")(y)
        y = keras.layers.Dense(classes, name="Head")(y)
    return keras.models.Model(inputs=x, outputs=y, name=name)


def vit_ti16(input_shape=(224, 224, 3),
             classes=1000,
             norm='LN',
             last_norm='LN',
             softmax='softmax',
             act='GeLU',
             include_top=True):
    """Build ViT-Tiny. All arguments passed to vit_imagenet.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        norm (str, optional): string that values in ['LN', 'GN1', 'BN', 'LMN'] and that allows to
            choose from LayerNormalization, GroupNormalization(groups=1, ...), BatchNormalization
            or LayerMadNormalization layers respectively in the model. Defaults to 'LN'.
        last_norm (str, optional): string that values in ['LN', 'BN']
            and that allows to choose from LayerNormalization or
            BatchNormalization in the classifier network. Defaults to 'LN'.
        softmax (str, optional): string with values in ['softmax', 'softmax2'] that allows to choose
            between softmax and softmax2 in attention block. Defaults to 'softmax'.
        act (str, optional): string that values in ['GeLU', 'ReLUx', 'swish'] and that allows to
            choose from GeLU, ReLUx or swish activation inside MLP. Defaults to 'GeLU'.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return vit_imagenet(
        name="vit-tiny",
        patch_size=16,
        input_shape=input_shape,
        classes=classes,
        norm=norm,
        last_norm=last_norm,
        act=act,
        softmax=softmax,
        include_top=include_top,
        **CONFIG_TI
    )


def bc_vit_ti16(input_shape=(224, 224, 3), classes=1000, include_top=True):
    """Build ViT-Tiny, changing all LN by LMN, using softmax2 and ReLU8.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return vit_imagenet(
        name="vit-tiny",
        patch_size=16,
        input_shape=input_shape,
        classes=classes,
        norm="LMN",
        last_norm="BN",
        softmax="softmax2",
        act="ReLU8",
        include_top=include_top,
        **CONFIG_TI
    )


def bc_vit_ti16_imagenet_pretrained():
    """ Helper method to retrieve a `bc_vit_ti16` model that was trained on ImageNet dataset.

    Returns:
        keras.Model, dict: a Keras Model instance and the quantization configuration as a JSON
        formatted dict.
    """
    model_name = 'bc_vit_ti16_224_i8_w4_a4.h5'
    file_hash = '1c2002c4cbc999290a12b25eeb4dd59660b0fd317082b395bd7116e764209f50'
    model_path = fetch_file(BASE_WEIGHT_PATH + model_name,
                            fname=model_name,
                            file_hash=file_hash,
                            cache_subdir='models')
    return load_model(model_path)


def vit_s16(input_shape=(224, 224, 3), classes=1000, include_top=True):
    """Build ViT-Small.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return vit_imagenet(
        name="vit-s16",
        patch_size=16,
        input_shape=input_shape,
        classes=classes,
        include_top=include_top,
        **CONFIG_S
    )


def vit_s32(input_shape=(224, 224, 3), classes=1000, include_top=True):
    """Build ViT-Small.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return vit_imagenet(
        name="vit-s32",
        patch_size=32,
        input_shape=input_shape,
        classes=classes,
        include_top=include_top,
        **CONFIG_S
    )


def vit_b16(input_shape=(224, 224, 3), classes=1000, include_top=True):
    """Build ViT-B16.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return vit_imagenet(
        name="vit-b16",
        patch_size=16,
        input_shape=input_shape,
        classes=classes,
        include_top=include_top,
        **CONFIG_B
    )


def vit_b32(input_shape=(224, 224, 3), classes=1000, include_top=True):
    """Build ViT-B32.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return vit_imagenet(
        name="vit-b32",
        patch_size=32,
        input_shape=input_shape,
        classes=classes,
        include_top=include_top,
        **CONFIG_B
    )


def vit_l16(input_shape=(384, 384, 3), classes=1000, include_top=True):
    """Build ViT-L16.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (384, 384, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return vit_imagenet(
        name="vit-l16",
        patch_size=16,
        input_shape=input_shape,
        classes=classes,
        include_top=include_top,
        **CONFIG_L
    )


def vit_l32(input_shape=(384, 384, 3), classes=1000, include_top=True):
    """Build ViT-L32.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (384, 384, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return vit_imagenet(
        name="vit-l32",
        patch_size=32,
        input_shape=input_shape,
        classes=classes,
        include_top=include_top,
        **CONFIG_L
    )



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Tools for calibration"""

import tensorflow as tf
from quantizeml.layers import Calibrable


def get_objects(model, object_type):
    def _get_objects(layer, objects):
        if isinstance(layer, object_type):
            objects.append(layer)
        for attr in layer.__dict__.values():
            if isinstance(attr, tf.keras.layers.Layer):
                _get_objects(attr, objects)
    objects = []
    for layer in model.layers:
        _get_objects(layer, objects)
    return objects


def calibration_required(model):
    """Checks if a model must be calibrated at training time.

    Args:
        model (keras.Model): the model to check

    Returns:
        bool: True if a calibration is required before a train or tune action.
    """
    calibrables = get_objects(model, Calibrable)
    if len(calibrables) == 0:
        return False

    for calibrable in calibrables:
        # If the model has never been calibrated, all max_value of the Calibrable objects
        # will be set to 1.
        if tf.reduce_any(calibrable.variables[0] != 1):
            return False
    # all calibrable objects are unset
    return True



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Data generator for CenterNet training
"""

import numpy as np
import tensorflow as tf
from imgaug import augmenters as iaa

from ..detection.batch_generator import BatchGenerator


class BatchCenternetGenerator(BatchGenerator):
    """
    Data generator used for the training process
    """
    @staticmethod
    def build_aug_pipeline():
        # augmentors by https://github.com/aleju/imgaug
        def sometimes(aug): return iaa.Sometimes(0.5, aug)

        # All augmenters with per_channel=0.5 will sample one value per
        # image in 50% of all cases. In all other cases they will sample new
        # values per channel.
        return iaa.Sequential(
            [
                # apply the following augmenters to most images
                sometimes(iaa.Affine()),
                sometimes(iaa.Cutout(nb_iterations=(0, 3), position="uniform", size=(0, 0.3))),
                # execute 0 to 5 of the following (less important) augmenters
                # per image. Don't execute all of them, as that would often be
                # way too strong
                iaa.SomeOf(
                    (0, 5),
                    [
                        iaa.OneOf([
                            # blur images with a sigma between 0 and 3.0
                            iaa.GaussianBlur((0, 3.0)),
                            # blur image using local means (kernel sizes between 2 and 7)
                            iaa.AverageBlur(k=(2, 7)),
                            # blur image using local medians (kernel sizes between 3 and 11)
                            iaa.MedianBlur(k=(3, 11)),
                        ]),
                        # sharpen images
                        iaa.Sharpen(alpha=(0, 1.0), lightness=(0.75, 1.5)),
                        iaa.LinearContrast((0.5, 2.0), per_channel=0.5),
                        iaa.AdditiveGaussianNoise(loc=0, scale=(0.0, 0.05 * 255), per_channel=0.5),
                        # randomly remove up to 10% of the pixels
                        iaa.OneOf([iaa.Dropout((0.01, 0.1), per_channel=0.5),
                                   iaa.CoarseDropout((0.01, 0.05), size_percent=0.5)]),
                        # change brightness of images
                        iaa.Add((-10, 10), per_channel=0.5),
                        iaa.Multiply((0.5, 1.5), per_channel=0.5),
                        iaa.OneOf([
                            iaa.pillike.Equalize(),
                            iaa.pillike.Autocontrast()
                        ])
                    ], random_order=True)
            ],
            random_order=True)

    def __getitem__(self, idx):
        lower_bound = idx * self._batch_size
        upper_bound = (idx + 1) * self._batch_size

        if upper_bound > self._data_length:
            upper_bound = self._data_length
            lower_bound = upper_bound - self._batch_size

        instance_count = 0

        N = upper_bound - lower_bound
        x_batch = np.zeros((N, self._input_shape[1], self._input_shape[0], self._input_shape[2]))
        y_batch = np.zeros((N, self._grid_size[1], self._grid_size[0], 2 + 2 + len(self._labels)))

        for train_instance in self._data[lower_bound:upper_bound]:
            # augment input image and fix object's position and size
            img, all_objs = self.aug_image(train_instance, jitter=self._jitter)

            for obj in all_objs:
                if obj['x2'] > obj['x1'] and obj['y2'] > obj['y1'] and obj['label'] in self._labels:
                    # find center
                    center_x = .5 * (obj['x1'] + obj['x2'])
                    # express it in output size
                    center_x = center_x / (float(self._input_shape[0]) / self._grid_size[0])
                    center_y = .5 * (obj['y1'] + obj['y2'])
                    center_y = center_y / (float(self._input_shape[1]) / self._grid_size[0])

                    grid_x = int(np.floor(center_x))
                    grid_y = int(np.floor(center_y))

                    if grid_x < self._grid_size[0] and grid_y < self._grid_size[1]:
                        obj_indx = self._labels.index(obj['label'])
                        scale_box_w = (obj['x2'] - obj['x1']) / \
                            (float(self._input_shape[0]) / self._grid_size[0])
                        scale_box_h = (obj['y2'] - obj['y1']) / (
                            float(self._input_shape[1]) / self._grid_size[0])

                        # get the center point and use a gaussian kernel as the target
                        radius = self._gaussian_radius([scale_box_h, scale_box_w], min_overlap=0.3)

                        # check that the radius is positive
                        radius = np.max(int(radius), 0)
                        heatmap = np.zeros(
                            (self._grid_size[0], self._grid_size[1], len(self._labels)))
                        heatmap = self._gen_gaussian_target(
                            heatmap, [grid_y, grid_x], obj_indx, radius)

                        # assign ground truth to y_batch
                        y_batch[instance_count, ..., :len(self._labels)] += heatmap
                        y_batch[instance_count, grid_y, grid_x, -4] = scale_box_w
                        y_batch[instance_count, grid_y, grid_x, -3] = scale_box_h
                        y_batch[instance_count, grid_y, grid_x, -2] = center_x - grid_x
                        y_batch[instance_count, grid_y, grid_x, -1] = center_y - grid_y

            # assign input image to x_batch
            x_batch[instance_count] = img

            # increase instance counter in current batch
            instance_count += 1

        return x_batch, y_batch

    @staticmethod
    def _gaussian2D(radius, sigma=1, eps=tf.keras.backend.epsilon()):
        """Generate 2D gaussian kernel.

        Args:
            radius (int): Radius of gaussian kernel.
            sigma (int, optional): Sigma of gaussian function. Defaults to 1.
            eps (float, optional): Epsilon value. Defaults to 1e-7.

        Returns:
            np.array: Gaussian kernel with a ``(2 * radius + 1) * (2 * radius + 1)`` shape.
        """
        x = np.reshape(np.arange(-radius, radius + 1), [1, -1])
        y = np.reshape(np.arange(-radius, radius + 1), [-1, 1])
        h = np.exp(-(x * x + y * y) / (2 * sigma * sigma))
        # Clamp smaller values to zero
        h[h < (eps * h.max())] = 0
        return h

    @staticmethod
    def _gen_gaussian_target(heatmap, center, obj_idx, radius):
        """Generate 2D gaussian heatmap.

        Args:
            heatmap (np.array): Input heatmap, the gaussian kernel will cover on it and maintain
                the max value.
            center (list[int]): Coordinates of gaussian kernel's center.
            obj_idx (int): The class index for the center point.
            radius (int): Radius of gaussian kernel.

        Returns:
            out_heatmap (np.array): Updated heatmap covered by gaussian kernel.

        Note:
            Taken from pytorch
        """
        diameter = 2 * radius + 1
        gaussian_kernel = BatchCenternetGenerator._gaussian2D(radius, sigma=diameter / 6)
        x, y = center
        height, width = heatmap.shape[:2]

        # Find the smallest value so that if the point is near the edge we don't end outside
        # (e.g. x = 3 and radius is 10, then we go from the x-3 to x+10)
        left = np.min([x, radius])
        right = np.min([width - x, radius + 1])
        top = np.min([y, radius])
        bottom = np.min([height - y, radius + 1])

        # Compare the gaussian kernel to the heatmap (in case there's already a point of
        # interest there) and keep the max value
        flattened_kernel = np.reshape(gaussian_kernel, [-1])

        # Range the dimensions
        i = 0
        for d0 in np.arange(y - top, y + bottom):
            for d1 in np.arange(x - left, x + right):
                heatmap[d1, d0, obj_idx] = flattened_kernel[i]
                i += 1
        return heatmap

    @staticmethod
    def _gaussian_radius(det_size, min_overlap):
        r"""Generate 2D gaussian radius.

        This function is modified from the `official github repo
        <https://github.com/princeton-vl/CornerNet-Lite/blob/master/core/sample/
        utils.py#L65>`_.

        Given ``min_overlap``, radius could computed by a quadratic equation
        according to Vieta's formulas.

        There are 3 cases for computing gaussian radius, details are following:

        - Case 1: one corner is inside the gt box and the other is outside.

        .. code:: text

            |<   width   >|

            lt-+----------+         -
            |  |          |         ^
            +--x----------+--+
            |  |          |  |
            |  |          |  |    height
            |  | overlap  |  |
            |  |          |  |
            |  |          |  |      v
            +--+---------br--+      -
            |          |  |
            +----------+--x

        To ensure IoU of generated box and gt box is larger than ``min_overlap``:

        .. math::
            \cfrac{(w-r)*(h-r)}{w*h+(w+h)r-r^2} \ge {iou} \quad\Rightarrow\quad
            {r^2-(w+h)r+\cfrac{1-iou}{1+iou}*w*h} \ge 0 \\
            {a} = 1,\quad{b} = {-(w+h)},\quad{c} = {\cfrac{1-iou}{1+iou}*w*h}
            {r} \le \cfrac{-b-\sqrt{b^2-4*a*c}}{2*a}

        - Case 2: both two corners are inside the gt box.

        .. code:: text

            |<   width   >|

            lt-+----------+         -
            |  |          |         ^
            +--x-------+  |
            |  |       |  |
            |  |overlap|  |       height
            |  |       |  |
            |  +-------x--+
            |          |  |         v
            +----------+-br         -

        To ensure IoU of generated box and gt box is larger than ``min_overlap``:

        .. math::
            \cfrac{(w-2*r)*(h-2*r)}{w*h} \ge {iou} \quad\Rightarrow\quad
            {4r^2-2(w+h)r+(1-iou)*w*h} \ge 0 \\
            {a} = 4,\quad {b} = {-2(w+h)},\quad {c} = {(1-iou)*w*h}
            {r} \le \cfrac{-b-\sqrt{b^2-4*a*c}}{2*a}

        - Case 3: both two corners are outside the gt box.

        .. code:: text

            |<   width   >|

            x--+----------------+
            |  |                |
            +-lt-------------+  |   -
            |  |             |  |   ^
            |  |             |  |
            |  |   overlap   |  | height
            |  |             |  |
            |  |             |  |   v
            |  +------------br--+   -
            |                |  |
            +----------------+--x

        To ensure IoU of generated box and gt box is larger than ``min_overlap``:

        .. math::
            \cfrac{w*h}{(w+2*r)*(h+2*r)} \ge {iou} \quad\Rightarrow\quad
            {4*iou*r^2+2*iou*(w+h)r+(iou-1)*w*h} \le 0 \\
            {a} = {4*iou},\quad {b} = {2*iou*(w+h)},\quad {c} = {(iou-1)*w*h} \\
            {r} \le \cfrac{-b+\sqrt{b^2-4*a*c}}{2*a}

        Args:
            det_size (list[int]): Shape of object.
            min_overlap (float): Min IoU with ground truth for boxes generated by
                keypoints inside the gaussian kernel.

        Returns:
            radius (int): Radius of gaussian kernel.

        Notes:
            Explanation of figure: ``lt`` and ``br`` indicates the left-top and bottom-right
            corner of ground truth box. ``x`` indicates the generated corner at the limited
            position when ``radius=r``.
        """
        height, width = det_size

        a1 = 1
        b1 = (height + width)
        c1 = width * height * (1 - min_overlap) / (1 + min_overlap)
        sq1 = np.sqrt(b1**2 - 4 * a1 * c1)
        r1 = (b1 - sq1) / (2 * a1)

        a2 = 4
        b2 = 2 * (height + width)
        c2 = (1 - min_overlap) * width * height
        sq2 = np.sqrt(b2**2 - 4 * a2 * c2)
        r2 = (b2 - sq2) / (2 * a2)

        a3 = 4 * min_overlap
        b3 = -2 * min_overlap * (height + width)
        c3 = (min_overlap - 1) * width * height
        sq3 = np.sqrt(b3**2 - 4 * a3 * c3)
        r3 = (b3 + sq3) / (2 * a3)

        return np.min([r1, r2, r3])



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for CenterNet models
"""
import os
import pickle
import argparse

from keras import Model

from ..training import get_training_parser, freeze_model_before, compile_model, save_model
from ..detection.map_evaluation import MapEvaluation
from ..param_scheduler import get_cosine_lr_scheduler
from ..utils import get_tensorboard_callback, load_model

from .centernet_batch_generator import BatchCenternetGenerator
from .centernet_loss import CenternetLoss
from .centernet_processing import decode_output


def get_data(data_path):
    """Loads data.

    Args:
        data_path (str): path to data pickle file

    Returns:
        dict, dict, list: train and validation data, labels
    """
    with open(data_path, 'rb') as handle:
        train_data, valid_data, labels = pickle.load(handle)
    return train_data, valid_data, labels


def _decode_centernet_output_fn(output, anchors, nb_classes, obj_threshold=0.5, nms_threshold=0.5):
    # Ignore attributes from original MAPEvaluator
    return decode_output(output, nb_classes, obj_threshold=obj_threshold)


def train(model, train_data, valid_data, labels, lr_max, obj_threshold,
          epochs, batch_size, grid_size, tune=False, out_dir=None):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train.
        train_data (dict): training data.
        valid_data (dict): validation data.
        labels (list): classes name.
        lr_max (float): max learning rate to take in learning rate scheduler.
        obj_threshold (float): confidence threshold for inference procedure.
        epochs (int): the number of epochs.
        batch_size (int): the batch size.
        grid_size (tuple): the spatial hw output size.
        tune (bool, optional): wheter the model will be tuned or not, modifying ``lr_max``.
            Defaults to False.
        out_dir (str, optional): folder name to save logs. Defaults to None
    """
    TRAIN_TIMES = 10

    # Build batch generators
    input_shape = model.input.shape[1:]

    # create the data generators
    train_generator = BatchCenternetGenerator(input_shape=input_shape,
                                              data=train_data,
                                              grid_size=grid_size,
                                              labels=labels,
                                              batch_size=batch_size)

    valid_generator = BatchCenternetGenerator(input_shape=input_shape,
                                              data=valid_data,
                                              grid_size=grid_size,
                                              labels=labels,
                                              batch_size=batch_size,
                                              jitter=False)

    # Create callbacks
    steps_per_epoch = len(train_generator) * TRAIN_TIMES
    map_evaluator_cb = MapEvaluation(model,
                                     valid_data,
                                     labels,
                                     anchors=None,
                                     period=4,
                                     obj_threshold=obj_threshold,
                                     decode_output_fn=_decode_centernet_output_fn)
    lrs_callback = get_cosine_lr_scheduler(lr_max,
                                           steps_per_epoch * epochs,
                                           pct_start=0.001 if tune else 0.3)
    tensorboard = get_tensorboard_callback(out_dir, prefix="centernet")

    callbacks = [map_evaluator_cb, lrs_callback, tensorboard]

    # Start the training process
    model.fit(x=train_generator,
              steps_per_epoch=steps_per_epoch,
              epochs=epochs,
              validation_data=valid_generator,
              callbacks=callbacks,
              workers=12,
              max_queue_size=40)


def evaluate(model, valid_data, labels, obj_threshold):
    """ Evaluates model performances.

    Args:
        model (keras.Model or akida.Model): the model to evaluate.
        valid_data (dict): evaluation data.
        labels (list): classes name.
        obj_threshold (float): confidence threshold for inference procedure.
    """
    # Create the mAP evaluator
    map_evaluator = MapEvaluation(model,
                                  valid_data,
                                  labels,
                                  anchors=None,
                                  obj_threshold=obj_threshold,
                                  is_keras_model=isinstance(model, Model),
                                  decode_output_fn=_decode_centernet_output_fn)

    # Compute mAP scores and display them
    mAP, average_precisions = map_evaluator.evaluate_map()
    for label, average_precision in average_precisions.items():
        print(labels[label], '{:.4f}'.format(average_precision))
    print('mAP: {:.4f}'.format(mAP))


def main():
    """ Entry point for script and CLI usage. """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument("-d", "--data", default=None,
                               help="The pickle file generated by the preprocessing step")
    global_parser.add_argument("-obj", "--obj_thresh", type=float, default=0.1,
                               help="Confidence threshold for a box. Defaults to %(default)s")
    global_parser.add_argument("-o", "--out_dir", type=str, default='./logs',
                               help="The output directory (logs). Defaults to %(default)s")
    parsers = get_training_parser(batch_size=128, freeze_before=True, tune=True,
                                  global_parser=global_parser)
    args = parsers[0].parse_args()

    # Load data
    train_data, valid_data, labels = get_data(args.data)

    # Load the source model
    model = load_model(args.model)

    grid_size = model.output.shape[1:3]

    nlabels, nmodel = len(labels), model.output.shape[-1]
    if nlabels + 4 != nmodel:
        raise ValueError(f"Model's output ({nmodel}) does not match with "
                         f"number of labels ({nlabels}) + 4. "
                         f"Check the data file {args.data} or input model {args.model}.")
    # Freeze the model
    if "freeze_before" in args:
        freeze_model_before(model, args.freeze_before)

    # Compile model
    learning_rate = 1e-4 if args.action == "tune" else 1e-2
    compile_model(model, learning_rate=learning_rate, loss=CenternetLoss(), metrics=None)

    # Disable QuantizeML assertions
    os.environ["ASSERT_ENABLED"] = "0"

    if args.action in ["train", "tune"]:
        train(model, train_data, valid_data, labels, learning_rate, args.obj_thresh,
              args.epochs, args.batch_size, grid_size, out_dir=args.out_dir,
              tune=args.action == "tune")
        save_model(model, args.model, args.savemodel, args.action)
    elif args.action == "eval":
        # Evaluate model accuracy
        # TODO: enable this feature when cnn2snn support SeparableConv2DTranspose
        # and full conversion into akida
        if args.akida:
            raise NotImplementedError("Not implemented yet!")
        evaluate(model, valid_data, labels, args.obj_thresh)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Akida U-Net model definition for semantic segmentation.
"""

from keras import Model, regularizers
from keras.layers import Dropout, Activation

from ..layer_blocks import sepconv_transpose_block, dense_block
from ..imagenet.model_akidanet import akidanet_imagenet


def akida_unet_portrait128(input_shape=(128, 128, 3),
                           alpha=0.5,
                           input_scaling=(128, -1)):
    """Instantiates an Akida U-Net architecture.

    It is composed of an AkidaNet-ImageNet encoder followed by a succession of Conv2DTranspose
    layers for the decoder part.
    It does not contain any skip connection (concatenation) between the encoder and the decoder
    branches.

    Args:
        input_shape (tuple, optional): input shape tuple. Defaults to (128, 128, 3).
        alpha (float, optional): controls the width (number of filters) of the model. Defaults to
            0.5.
        input_scaling (tuple, optional): scale factor and offset to apply to inputs. Defaults to
            (128, -1). Note that following Akida convention, the scale factor is a number used as a
            divider.

    Returns:
        keras.Model: a Keras Model instance.
    """
    # Define weight regularization, will apply to pointwise weights of sepconv transposed layers
    weight_regularizer = regularizers.l2(4e-5)

    # Create an AkidaNet network without top layers
    encoder = akidanet_imagenet(input_shape=input_shape,
                                alpha=alpha,
                                include_top=False,
                                input_scaling=input_scaling)

    # Add the decoder layers
    x = encoder.layers[-1].output
    x = sepconv_transpose_block(x,
                                filters=int(512 * alpha),
                                kernel_size=(3, 3),
                                strides=2,
                                padding='same',
                                kernel_initializer='he_normal',
                                add_batchnorm=True,
                                name='sepconv_t_0',
                                pointwise_regularizer=weight_regularizer)
    x = Dropout(0.5)(x)
    x = sepconv_transpose_block(x,
                                filters=int(256 * alpha),
                                kernel_size=(3, 3),
                                strides=2,
                                padding='same',
                                kernel_initializer='he_normal',
                                add_batchnorm=True,
                                name='sepconv_t_1',
                                pointwise_regularizer=weight_regularizer)
    x = Dropout(0.5)(x)
    x = sepconv_transpose_block(x,
                                filters=int(128 * alpha),
                                kernel_size=(3, 3),
                                strides=2,
                                padding='same',
                                kernel_initializer='he_normal',
                                add_batchnorm=True,
                                name='sepconv_t_2',
                                pointwise_regularizer=weight_regularizer)
    x = Dropout(0.5)(x)
    x = sepconv_transpose_block(x,
                                filters=int(64 * alpha),
                                kernel_size=(3, 3),
                                strides=2,
                                padding='same',
                                kernel_initializer='he_normal',
                                add_batchnorm=True,
                                name='sepconv_t_3',
                                pointwise_regularizer=weight_regularizer)
    x = Dropout(0.5)(x)
    x = sepconv_transpose_block(x,
                                filters=int(32 * alpha),
                                kernel_size=(3, 3),
                                strides=2,
                                padding='same',
                                kernel_initializer='he_normal',
                                add_batchnorm=True,
                                name='sepconv_t_4',
                                pointwise_regularizer=weight_regularizer)
    x = Dropout(0.5)(x)
    x = dense_block(x, units=1, add_activation=False, name='head')
    x = Activation('sigmoid', name="sigmoid_act")(x)

    # Build the whole model: encoder followed by decoder
    return Model(inputs=encoder.input, outputs=x, name='akida_unet')



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
CenterNet model definition for detection
"""
import numpy as np

from keras import Model, initializers
from keras.layers import Conv2D


from .. import akidanet_imagenet
from ..layer_blocks import (separable_conv_block, conv_block, conv_transpose_block,
                            sepconv_transpose_block)


def centernet_base(input_shape=(224, 224, 3),
                   classes=2,
                   alpha=0.5,
                   input_scaling=(127, -1)):
    """ A Keras Model implementing the CenterNet architecture, on top of an AkidaNet backbone

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of output classes. Defaults to 2.
        alpha (float): controls the width of the model. Defaults to 0.5.
        input_scaling (tuple, optional): input scaling. Defaults to (127, -1).

    Returns:
        keras.Model: a Keras Model instance.
    """
    # Create an AkidaNet network without top layers
    base_model = akidanet_imagenet(
        input_shape=input_shape,
        alpha=alpha,
        include_top=False,
        input_scaling=input_scaling)

    x = base_model.layers[-1].output

    # Build the neck with up convolutions
    num_deconv_filters = [256, 128, 64]

    for i, (n_filt) in enumerate(num_deconv_filters):
        if n_filt <= 128:
            curr_block = conv_block
            block_type = "conv"
            transpose_block = conv_transpose_block
        else:
            curr_block = separable_conv_block
            block_type = "sepconv"
            transpose_block = sepconv_transpose_block

        x = curr_block(x,
                       filters=n_filt,
                       name=f'neck_{block_type}_{i}',
                       kernel_size=(3, 3),
                       padding='same',
                       use_bias=False,
                       add_activation=True,
                       add_batchnorm=True)
        x = transpose_block(x,
                            filters=n_filt,
                            name=f"neck_transpose_{block_type}_{i}",
                            kernel_size=(4, 4),
                            padding="same",
                            strides=(2, 2),
                            use_bias=False,
                            add_activation=True,
                            add_batchnorm=True)

    # Build the head which is composed of 2 consecutive convs
    bias_initializer = initializers.Constant(float(-np.log((1 - 0.1) / 0.1)))
    init_kernel = initializers.RandomNormal(stddev=0.001, seed=6)
    # In the legacy model there is 3 branches of 64 filters each one.
    # This could be merged in one branch
    x = conv_block(x, 3 * 64, (3, 3),
                   add_batchnorm=True,
                   use_bias=False,
                   add_activation=True,
                   padding="same",
                   name="head_conv_1",
                   kernel_initializer=init_kernel)
    # The output is built by #classes and box coordinates in xywh
    x = Conv2D(classes + 4,
               (1, 1),
               padding="same",
               use_bias=True,
               name="head_conv_2",
               bias_initializer=bias_initializer,
               kernel_initializer=init_kernel)(x)

    # Build the model
    return Model(inputs=base_model.input, outputs=x, name='centernet_base')
